# 2010 October 5 by Joshua R. Davis

import OpenGL.GL as gl
import math
from utility import *

class Body2D(object):
	"""A 2D rigid body that can draw a 2D mesh with rotation, translation, and scaling, and with various combinations of color and texture."""
	
	def __init__(self, mesh=None, textures=[], uniforms=[]):
		"""Initializes the body with a Mesh, a list of Textures, and a list of uniform properties."""
		self.setPosition((0.0, 0.0))
		self.setAngle(0.0)
		self.setScales((1.0, 1.0))
		self.uniforms = uniforms[:]
		self.textures = textures[:]
		self.setMesh(mesh)
		self.setResizingBehavior("start", "start")
	
	def getExampleBody(self):
		"""Non-user method used by Renderer2D. If this body is ready to draw, as indicated by having a mesh, then returns self. Otherwise returns None."""
		if self.mesh == None:
			return None
		else:
			return self
	
	def setMesh(self, mesh):
		"""Sets the mesh. The user is responsible for its being compatible with the renderer."""
		self.mesh = mesh
	
	def getMesh(self):
		"""Returns the mesh."""
		return self.mesh
	
	def getTextures(self):
		"""Returns the list of textures, which the user can alter (append, pop, etc.) as desired, as long as they remain compatible with the renderer."""
		return self.textures
	
	def getUniforms(self):
		"""Returns the list of uniform properties, which the user can alter (append, pop, etc.) as desired, as long as they remain compatible with the renderer."""
		return self.uniforms
	
	def setPosition(self, xy):
		"""Sets the (x, y) position."""
		self.position = xy
	
	def getPosition(self):
		"""Returns the (x, y) position."""
		return self.position
	
	def setAngle(self, angle):
		"""Controls the orientation by specifying a rotation angle (measured counterclockwise from the positive x-axis in radians)."""
		self.angle = angle
	
	def getAngle(self):
		"""Returns the orientation angle."""
		return self.angle
	
	def setScales(self, xy):
		"""Sets the x- and y-scales to the given pair xy of floats. The default value is (1.0, 1.0)."""
		self.scales = tuple(xy)
	
	def getScales(self):
		"""Returns the x- and y-scales as a pair of floats."""
		return self.scales
	
	def translate(self, v):
		"""Changes the position by adding v = (dx, dy)."""
		self.position = (self.position[0] + v[0], self.position[1] + v[1])
	
	def rotate(self, angle):
		"""Rotates the body counterclockwise by the given angle (in radians)."""
		self.angle += angle
	
	def revolve(self, angle, center):
		"""Revolves the body about the given center point by the given angle; changes position but not orientation."""
		delta = (self.position[0] - center[0], self.position[1] - center[1])
		cosine = math.cos(angle)
		sine = math.sin(angle)
		delta = (cosine * delta[0] - sine * delta[1], sine * delta[0] + cosine * delta[1])
		self.position = (center[0] + delta[0], center[1] + delta[1])
	
	def revolveRotate(self, angle, center):
		"""Revolves the body about the given center point by the given angle, also adding this angle to the body's angle, so that the same side of the body always faces the center."""
		self.revolve(angle, center)
		self.rotate(angle)
	
	def getMatrix(self):
		"""Non-user. Returns the product of the receiver's translation, rotation, and scaling matrices. This matrix takes local object coordinates to world coordinates."""
		cosine = math.cos(self.angle)
		sine = math.sin(self.angle)
		return (cosine * self.scales[0], -sine * self.scales[1], 0.0, self.position[0], sine * self.scales[0], cosine * self.scales[1], 0.0, self.position[1], 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0)
	
	def draw(self, uniformLocations, attributeLocations, matrixLocation, m=None):
		"""Non-user. Draws the body using its renderer's locations. The uniform locations correspond one-to-one to the body's uniforms and the attribute locations correspond one-to-one to the mesh's attributes. Any location that is -1 is ignored. matrixLocation is the location of the 4x4 modeling matrix; m is the modeling matrix established by an enclosing group, or None. Matrices are stored as length-16 vectors in row-major order."""
		if self.mesh != None:
			# Establish the modeling transformation.
			if m == None:
				gl.glUniformMatrix4fv(matrixLocation, 1, gl.GL_TRUE, self.getMatrix())
			else:
				gl.glUniformMatrix4fv(matrixLocation, 1, gl.GL_TRUE, affineMatrixProduct(m, self.getMatrix()))
			# Set the textures and uniforms.
			for i in range(len(self.textures)):
				gl.glActiveTexture(gl.GL_TEXTURE0 + i)
				self.textures[i].bind()
			for i in range(len(uniformLocations)):
				if uniformLocations[i] != -1:
					if len(self.uniforms[i]) == 4:
						gl.glUniform4f(uniformLocations[i], self.uniforms[i][0], self.uniforms[i][1], self.uniforms[i][2], self.uniforms[i][3])
					elif len(self.uniforms[i]) == 3:
						gl.glUniform3f(uniformLocations[i], self.uniforms[i][0], self.uniforms[i][1], self.uniforms[i][2])
					elif len(self.uniforms[i]) == 2:
						gl.glUniform2f(uniformLocations[i], self.uniforms[i][0], self.uniforms[i][1])
					else:
						gl.glUniform1f(uniformLocations[i], self.uniforms[i][0])
			# Draw the mesh.
			self.mesh.draw(attributeLocations)
	
	def setResizingBehavior(self, xBehavior, yBehavior):
		"""Sets the resizing behavior in the x- and y-directions. For each, the options are 'start' (hold position relative to the left or bottom side), 'end' (hold position relative to the right or top side), or 'fraction' (maintain the fraction of the way between the two sides)."""
		self.xBehavior = xBehavior
		self.yBehavior = yBehavior
	
	def getResizingBehavior(self):
		"""Returns the resizing behavior (xBehavior, yBehavior) in the x- and y-directions."""
		return (self.xBehavior, self.yBehavior)
	
	def setResizingWidthHeight(self, oldWidth, oldHeight, newWidth, newHeight):
		"""Non-user. Method called by Renderer2D whenever the window changes size. Does nothing. Subclasses may override without calling this version."""
		if self.xBehavior == "end":
			x = self.position[0] + newWidth - oldWidth
		elif self.xBehavior == "fraction":
			x = float(self.position[0]) / oldWidth * newWidth
		else:
			x = self.position[0]
		if self.yBehavior == "end":
			y = self.position[1] + newHeight - oldHeight
		elif self.yBehavior == "fraction":
			y = float(self.position[1]) / oldHeight * newHeight
		else:
			y = self.position[1]
		self.position = (x, y)

class Group2D(Body2D):
	"""A body that does no drawing of its own, but rather collects other bodies into a group, so that they act as if they're glued together. All of the collected bodies are drawn with their own colors and textures, but with position, orientation, and scaling all relative to the group's."""
	
	def __init__(self, bodies=[]):
		"""Initializes the group with the given list of bodies."""
		Body2D.__init__(self)
		self.bodies = bodies[:]
	
	def getBodies(self):
		"""Returns the list of bodies in the group, which the user can alter (append, pop, etc.) as desired, as long as they remain compatible with the renderer."""
		return self.bodies
	
	def getExampleBody(self):
		"""Non-user method used by Renderer2D. Overrides to return an exemplary body from a child, if there isn't one in the receiver."""
		example = Body2D.getExampleBody(self)
		if example == None:
			for body in self.bodies:
				example = body.getExampleBody()
				if example != None:
					return example
			return None
		else:
			return example
	
	def draw(self, uniformLocations, attributeLocations, matrixLocation, m=None):
		"""Non-user. Overrides Body2D's version to draw the group's bodies relative to the group."""
		if m == None:
			matrix = self.getMatrix()
		else:
			matrix = affineMatrixProduct(m, self.getMatrix())
		for body in self.bodies:
			if body != None:
				body.draw(uniformLocations, attributeLocations, matrixLocation, matrix)
