# 2010 July 9 by Joshua R. Davis

import OpenGL.GL as gl
import OpenGL.GL.EXT.framebuffer_object as fbo
import ode
import math
import window
from utility import *

class Camera(object):
	"""An Camera is a frustum or box that describes the viewing volume of a perspective or orthogonal camera (or the field of illumination of a spot or directional light). To the user, a camera appears to be an actual camera situated in world space, with definite position, orientation, and lens characteristics. The camera's position and orientation are controlled by an underlying Open Dynamics Engine Body. All messages that the camera receives and doesn't recognize are delegated to the ODE Body. In most cases this means that you can talk to Camera as if it's a subclass of ODE's Body. For example, you can set the camera's position by calling myBody3D.setPosition(xyz) instead of myBody3D.getBody().setPosition(xyz)."""
	
	def __init__(self, isOrthogonal=False, world=None):
		"""Initializes the camera to its default settings. The camera is either perspective or orthogonal; this cannot be changed. The world argument should be an ODE World object, where the body will live; if world is None, then the body lives in the world of the current window."""
		if world == None:
			world = window.getWindow().getWorld()
		# This is like self.body = ode.Body(world), but doesn't trigger __setattr__.
		self.__dict__["body"] = ode.Body(world)
		self.isOrthogonal = isOrthogonal
		self.setNearFar(1.0, 100.0)
		# Set the aspect ratio manually.
		self.ratio = 1.0
		self.setAngle(math.pi / 6.0)
		self.getBody().setPosition((0.0, 0.0, 10.0))
	
	def __getattr__(self, attr):
		"""Non-user method; part of the system for delegating unrecognized messages to the underlying ODE Body."""
		return getattr(self.body, attr)
	
	def __setattr__(self, attr, value):
		"""Non-user method; part of the system for delegating unrecognized messages to the underlying ODE Body."""
		return setattr(self.body, attr, value)
	
	def getBody(self):
		"""Returns the underlying ODE Body, which is never None. Note however that in most cases you do not need to get the ODE Body; all messages that the Camera receives, that it doesn't recognize, are delegated to the ODE Body automatically."""
		return self.body
	
	def translate(self, v):
		"""Changes the position by adding v = (dx, dy, dz)."""
		self.body.setPosition(vectorSum(self.body.getPosition(), v))
	
	def rotate(self, angle, axis):
		"""Prepends the given angle-axis rotation to the receiver's orientation."""
		old = self.body.getQuaternion()
		new = quaternionFromAngleAxis((angle, axis))
		product = quaternionProduct(new, old)
		self.body.setQuaternion(product)
	
	def revolve(self, angle, axis, center):
		"""Revolves the object about the given center point by the angle-axis rotation; changes position but not orientation."""
		delta = vectorDifference(self.body.getPosition(), center)
		delta = vectorRotated((angle, axis), delta)
		self.body.setPosition(vectorSum(center, delta))
	
	def revolveRotate(self, angle, axis, center):
		"""Revolves the object about the given center point by the given angle-axis rotation, also prepending the rotation to the object's orientation. Useful for keeping a camera or light pointed at an object as it revolves about it."""
		self.revolve(angle, axis, center)
		self.rotate(angle, axis)
	
	def dolly(self, distance):
		"""Moves the camera forward by distance (or backward, if negative). In orthogonal mode, this has no visual effect."""
		v = self.getBody().vectorToWorld((0.0, 0.0, -1.0))
		v = vectorScaling(v, distance)
		self.getBody().setPosition(vectorSum(self.body.getPosition(), v))
	
	def getIsOrthogonal(self):
		"""Returns whether the camera is orthogonal or not (perspective)."""
		return self.isOrthogonal
	
	def setNearFar(self, near, far):
		"""Controls world-space distance to near and far clipping planes. The user must keep far > near. The defaults are 1 and 100."""
		self.near = near
		self.far = far
	
	def getNear(self):
		"""Returns the distance to the near clipping plane, in world space."""
		return self.near
	
	def getFar(self):
		"""Returns the distance to the far clipping plane, in world space."""
		return self.far
	
	def setAspectRatio(self, ratio):
		"""Sets the aspect ratio (width / height)."""
		self.ratio = ratio
		self.halfWidth = self.ratio * self.halfHeight
	
	def getAspectRatio(self):
		"""Returns the aspect ratio (width / height)."""
		return self.ratio
	
	def setHeight(self, height):
		"""Sets the height of the viewing volume (on the near plane)."""
		self.halfHeight = 0.5 * height
		self.halfAngle = math.atan(self.halfHeight / self.near)
		self.halfWidth = self.ratio * self.halfHeight
	
	def getHeight(self):
		"""Returns the height of the viewing volume (on the near plane)."""
		return 2.0 * self.halfHeight
	
	def setAngle(self, angle):
		"""Sets the perspective frustum angle (not the half angle) in radians."""
		self.halfAngle = 0.5 * angle
		self.halfHeight = self.near * math.tan(self.halfAngle)
		self.halfWidth = self.ratio * self.halfHeight
	
	def getAngle(self):
		"""Returns the perspective frustum angle (not the half angle) in radians."""
		return 2.0 * self.halfAngle
	
	def getMatrix(self):
		"""Non-user. Returns the 4x4 matrix that expresses the camera's viewing transformations. It is the product of a projection, an inverse rotation, and an inverse translation. It does not include the viewport transformation or the various modeling transformations for the various objects in the scene."""
		# Get the projection matrix.
		if self.isOrthogonal:
			p = (1.0 / self.halfWidth, 0.0, 0.0, 0.0, 0.0, 1.0 / self.halfHeight, 0.0, 0.0, 0.0, 0.0, 2.0 / (self.near - self.far), (self.near + self.far) / (self.near - self.far), 0.0, 0.0, 0.0, 1.0)
		else:
			p = (self.near / self.halfWidth, 0.0, 0.0, 0.0, 0.0, self.near / self.halfHeight, 0.0, 0.0, 0.0, 0.0, (self.near + self.far) / (self.near - self.far), 2.0 * self.near * self.far / (self.near - self.far), 0.0, 0.0, -1.0, 0.0)
		# Get the inverse rotation as the 4x4 version of the transpose of the ODE rotation matrix.
		rot = self.body.getRotation()
		t = [rot[0 * 3 + 0], rot[1 * 3 + 0], rot[2 * 3 + 0], 0.0, rot[0 * 3 + 1], rot[1 * 3 + 1], rot[2 * 3 + 1], 0.0, rot[0 * 3 + 2], rot[1 * 3 + 2], rot[2 * 3 + 2], 0.0, 0.0, 0.0, 0.0, 1.0]
		# Multiply on the inverse translation; this changes just the last column.
		pos = self.body.getPosition()
		t[0 * 4 + 3] = -(t[0 * 4 + 0] * pos[0] + t[0 * 4 + 1] * pos[1] + t[0 * 4 + 2] * pos[2])
		t[1 * 4 + 3] = -(t[1 * 4 + 0] * pos[0] + t[1 * 4 + 1] * pos[1] + t[1 * 4 + 2] * pos[2])
		t[2 * 4 + 3] = -(t[2 * 4 + 0] * pos[0] + t[2 * 4 + 1] * pos[1] + t[2 * 4 + 2] * pos[2])
		# Multiply p * t, using the fact that in p only the 00-, 11-, 22-, 23-, 32-, and 33-entries can be nonzero.
		return (p[0 * 4 + 0] * t[0 * 4 + 0], p[0 * 4 + 0] * t[0 * 4 + 1], p[0 * 4 + 0] * t[0 * 4 + 2], p[0 * 4 + 0] * t[0 * 4 + 3], p[1 * 4 + 1] * t[1 * 4 + 0], p[1 * 4 + 1] * t[1 * 4 + 1], p[1 * 4 + 1] * t[1 * 4 + 2], p[1 * 4 + 1] * t[1 * 4 + 3], p[2 * 4 + 2] * t[2 * 4 + 0], p[2 * 4 + 2] * t[2 * 4 + 1], p[2 * 4 + 2] * t[2 * 4 + 2], p[2 * 4 + 2] * t[2 * 4 + 3] + p[2 * 4 + 3], p[3 * 4 + 2] * t[2 * 4 + 0], p[3 * 4 + 2] * t[2 * 4 + 1], p[3 * 4 + 2] * t[2 * 4 + 2], p[3 * 4 + 2] * t[2 * 4 + 3] + p[3 * 4 + 3])
	
	def getInverseMatrix(self):
		"""Non-user. Returns the 4x4 matrix that expresses the inverse of the camera's viewing transformations. It is the product of a translation, a rotation, and an inverse projection. It does not include the viewport transformation or the various modeling transformations for the various objects in the scene."""
		# Get the inverse projection matrix.
		if self.isOrthogonal:
			p = (self.halfWidth, 0.0, 0.0, 0.0, 0.0, self.halfHeight, 0.0, 0.0, 0.0, 0.0, (self.near - self.far) / 2.0, -0.5 * (self.near + self.far), 0.0, 0.0, 0.0, 1.0)
		else:
			p = (self.halfWidth / self.near, 0.0, 0.0, 0.0, 0.0, self.halfHeight / self.near, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 0.5 * (self.near - self.far) / (self.far * self.near), 0.5 * (self.near + self.far) / (self.far * self.near))
		# Get the translation times the rotation.
		rot = self.body.getRotation()
		pos = self.body.getPosition()
		t = [rot[0 * 3 + 0], rot[0 * 3 + 1], rot[0 * 3 + 2], pos[0], rot[1 * 3 + 0], rot[1 * 3 + 1], rot[1 * 3 + 2], pos[1], rot[2 * 3 + 0], rot[2 * 3 + 1], rot[2 * 3 + 2], pos[2], 0.0, 0.0, 0.0, 1.0]
		# Multiply t * p, using the fact that in p only the 00-, 11-, 22-, 23-, 32-, and 33-entries can be nonzero.
		return (t[0 * 4 + 0] * p[0 * 4 + 0], t[0 * 4 + 1] * p[1 * 4 + 1], t[0 * 4 + 2] * p[2 * 4 + 2] + t[0 * 4 + 3] * p[3 * 4 + 2], t[0 * 4 + 2] * p[2 * 4 + 3] + t[0 * 4 + 3] * p[3 * 4 + 3], t[1 * 4 + 0] * p[0 * 4 + 0], t[1 * 4 + 1] * p[1 * 4 + 1], t[1 * 4 + 2] * p[2 * 4 + 2] + t[1 * 4 + 3] * p[3 * 4 + 2], t[1 * 4 + 2] * p[2 * 4 + 3] + t[1 * 4 + 3] * p[3 * 4 + 3], t[2 * 4 + 0] * p[0 * 4 + 0], t[2 * 4 + 1] * p[1 * 4 + 1], t[2 * 4 + 2] * p[2 * 4 + 2] + t[2 * 4 + 3] * p[3 * 4 + 2], t[2 * 4 + 2] * p[2 * 4 + 3] + t[2 * 4 + 3] * p[3 * 4 + 3], 0.0, 0.0, p[3 * 4 + 2], p[3 * 4 + 3])
	
	def project(self, xyz):
		"""Projects a world point to a screen point, using the camera's state as well as the OpenGL viewport. That is, returns a tuple (x, y, z) of screen coordinates for the given world coordinates. This method is the inverse of unproject()."""
		viewport = gl.glGetIntegerv(gl.GL_VIEWPORT)
		halfWidth = 0.5 * viewport[2]
		halfHeight = 0.5 * viewport[3]
		v = matrixProduct(self.getMatrix(), (xyz[0], xyz[1], xyz[2], 1.0), 4, 4, 1)
		return ((v[0] / v[3] + 1.0) * halfWidth, (v[1] / v[3] + 1.0) * halfHeight, (v[2] / v[3] + 1.0) * 0.5)
	
	def unproject(self, xyz):
		"""Unprojects a screen point to a world point, using the camera's state as well as the OpenGL viewport. That is, returns a tuple (x, y, z) of world coordinates for the given screen coordinates. The screen coordinates are typically obtained by reading a depth value out of a particular pixel. This method is the inverse of project()."""
		viewport = gl.glGetIntegerv(gl.GL_VIEWPORT)
		halfWidth = 0.5 * viewport[2]
		halfHeight = 0.5 * viewport[3]
		w = (xyz[0] / halfWidth - 1.0, xyz[1] / halfHeight - 1.0, xyz[2] * 2.0 - 1.0, 1.0)
		v = matrixProduct(self.getInverseMatrix(), w, 4, 4, 1)
		return (v[0] / v[3], v[1] / v[3], v[2] / v[3])
	
	def getDirection(self):
		"""Returns the unit 3-vector in the direction that the camera is pointing."""
		return self.body.vectorToWorld((0.0, 0.0, -1.0))
	
	def getPickLine(self, xy):
		"""Returns a pair (p, v), where p is a point and v is a unit vector, such that p + t v parametrizes the line through the scene that passes through screen coordinates xy (i.e. (x, y, 0.5))."""
		xyz = self.unproject((xy[0], xy[1], 0.5))
		if self.isOrthogonal:
			return [xyz, self.getBody().vectorToWorld((0.0, 0.0, -1.0))]
		else:
			pos = self.getBody().getPosition()
			return [pos, unitVector(vectorDifference(xyz, pos))]
	
	def moveScreenPoint(self, xyPrevious, xyCurrent):
		"""This is a simple way to implement Window's member drag handler so that the mouse controls the camera. The camera adjusts so that what used to appear at the previous xy screen coordinates now appears at the current xy screen coordinates. In perspective mode, it does this by rotating the camera; in orthogonal mode, it translates the camera in a direction perpendicular to the line of sight."""
		if self.isOrthogonal:
			# Get displacement in world coordinates.
			dx = xyPrevious[0] - xyCurrent[0]
			dy = xyPrevious[1] - xyCurrent[1]
			# Get transformed x and y unit vectors.
			vx = self.getBody().vectorToWorld((1.0, 0.0, 0.0))
			vy = self.getBody().vectorToWorld((0.0, 1.0, 0.0))
			# Translate the camera.
			vx = vectorScaling(vx, dx)
			vy = vectorScaling(vy, dy)
			pos = vectorSum(vectorSum(self.body.getPosition(), vx), vy)
			self.getBody().setPosition(pos)
		else:
			# Get previous and current mouse directions.
			dirPrev = self.getPickLine(xyPrevious)[1]
			dirCurr = self.getPickLine(xyCurrent)[1]
			# Rotate the camera.
			quat = quaternionFromVectors(dirCurr, dirPrev)
			self.getBody().setQuaternion(quaternionProduct(quat, self.getBody().getQuaternion()))
	
	def getUpwardCorrection(self):
		"""Returns the angle-axis rotation about the camera's sightline that, when prepended to the camera's current orientation, makes the camera upright. If this calculation is impossible (because the camera's sightline is vertical) then None is returned."""
		# Get the local forward- and upward-pointing vectors.
		f = self.getBody().vectorToWorld((0.0, 0.0, -1.0))
		u = self.getBody().vectorToWorld((0.0, 1.0, 0.0))
		# Get the most upward-pointing vector perpendicular to the forward one.
		z = 1.0 - f[2] * f[2]
		if z <= 0.0:
			return None
		else:
			z = math.sqrt(z)
			if z == 0.0:
				return None
			else:
				newu = (-f[0] * f[2] / z, -f[1] * f[2] / z, z)
				return [angleFromAxisAndVectors(f, u, newu), f]
	
	def containPoints(self, points):
		"""Given a sequence of points in global coordinates, adjusts the viewing volume so that it contains all of the points. The convex hull of the points should have positive volume. May change any of near, far, aspect ratio, height, and angle. Returns Boolean indicating success. For orthogonal cameras, always succeeds but may translate the camera. For perspective cameras, never translates the camera but sometimes fails --- namely, if any point is behind the camera."""
		# Transform the points to local coordinates.
		locals = [self.body.getPosRelPoint(p) for p in points]
		if self.isOrthogonal:
			# Compute an axis-aligned bounding box for the points.
			minX = min([p[0] for p in locals])
			maxX = max([p[0] for p in locals])
			minY = min([p[1] for p in locals])
			maxY = max([p[1] for p in locals])
			minZ = min([p[2] for p in locals])
			maxZ = max([p[2] for p in locals])
			# Choose the aspect ratio and height.
			width = maxX - minX
			height = maxY - minY
			self.setAspectRatio(width / height)
			self.setHeight(height)
			# Translate to center the camera on the viewing box.
			x = 0.5 * (maxX + minX)
			y = 0.5 * (maxY + minY)
			self.translate(self.body.vectorToWorld((x, y, 0.0)))
			self.setNearFar(-maxZ, -minZ)
			return True
		else:
			# Find the most extreme values of the x- and y-tangents and z.
			xTans = []
			yTans = []
			zs = []
			for p in locals:
				if p[2] <= 0.0:
					return False
				else:
					xTans.append(math.fabs(p[0] / p[2]))
					yTans.append(math.fabs(p[1] / p[2]))
					zs.append(p[2])
			# Set the near and far planes, and use them to set height and aspect ratio.
			self.setNearFar(-max(zs), -min(zs))
			width = self.near * max(xTans)
			height = self.near * max(yTans)
			self.setAspectRatio(width / height)
			self.setHeight(height)
			return True
	
	def getLocalVertices(self):
		"""Returns a list of the 8 vertices of the viewing volume in local coordinates."""
		near = [(-self.halfWidth, -self.halfHeight, -self.near), (self.halfWidth, -self.halfHeight, -self.near), (self.halfWidth, self.halfHeight, -self.near), (-self.halfWidth, self.halfHeight, -self.near)]
		if self.isOrthogonal:
			far = [(p[0], p[1], -self.far) for p in near]
		else:
			ratio = self.far / self.near
			far = [vectorScaling(p, ratio) for p in near]
		return near + far
	
	def getVertices(self):
		"""Returns a list of the 8 vertices of the frustum as points in world space."""
		return [self.body.getRelPointPos(p) for p in self.getLocalVertices()]

class Light(Camera):
	"""A point light source positioned somewhere in the scene, that casts light (of some specified RGB color) in all directions. The orientation of the light in space therefore has no visual effect. This class does not support shadow-casting. The orthogonality of the underlying camera has no effect on the light."""
	
	def __init__(self, world=None, isOrthogonal=False):
		"""Initializes the light. The isOrthogonal argument is passed along to Camera.__init__; it indicates that point lights are not orthogonal."""
		Camera.__init__(self, isOrthogonal, world)
		self.color = (1.0, 1.0, 1.0)
	
	def getIsDirectional(self):
		"""Non-user. Invoked by Renderer3D while building its shader programs."""
		return self.getIsOrthogonal()
	
	def setColor(self, rgb):
		"""Sets the RGB color of the light to the given tuple. The default is (1.0, 1.0, 1.0)."""
		self.color = tuple(rgb)
	
	def getColor(self):
		"""Returns the RGB color as a 3-tuple."""
		return self.color
	
	def getUniforms(self):
		"""Non-user method used by Renderer3D during the construction of its shader program. Returns a tuple of the uniform properties of the light. These properties are set in other methods."""
		return ((self.color[0], self.color[1], self.color[2], 1.0),)
	
	def getCastsShadows(self):
		"""Returns False; shadowing of omnidirectional point lights is not supported."""
		return False
	
	def prepareShadows(self):
		"""Non-user. Does nothing; shadowing of omnidirectional point lights is not supported."""
		pass
	
	def draw(self, uniformLocations):
		"""Non-user method called by Renderer3D during the drawing of the scene. The light sets its uniform values into the given GLSL uniform locations (except the locations that are -1)."""
		uniforms = self.getUniforms()
		for i in range(len(uniformLocations)):
			if uniformLocations[i] != -1:
				if len(uniforms[i]) == 4:
					gl.glUniform4f(uniformLocations[i], uniforms[i][0], uniforms[i][1], uniforms[i][2], uniforms[i][3])
				elif len(uniforms[i]) == 3:
					gl.glUniform3f(uniformLocations[i], uniforms[i][0], uniforms[i][1], uniforms[i][2])
				elif len(uniforms[i]) == 2:
					gl.glUniform2f(uniformLocations[i], uniforms[i][0], uniforms[i][1])
				else:
					gl.glUniform1f(uniformLocations[i], uniforms[i][0])
	
	def getShaderCode(self, lightIndex):
		"""Non-user method called by Renderer3D during construction of the shader program. Returns a tuple of four strings of GLSL code, any of them possibly empty. The first string contains statements to be incorporated into the fragment shader program main body. The second string contains an expression to be and-ed into the diffuse/specular criterion for the light. The third string contains statements to be incorporated into the code that is executed only for those fragments that the light shines upon. The fourth string contains an expression to be multiplied into the amount of light reaching the fragment. Any variable names introduced should begin with 'lightI', where 'I' is the lightIndex (a small nonnegative integer). The light's uniforms can be accessed in this code as 'lIuJ', where 'I' is the lightIndex and 'J' is the index of the uniform in the light's uniform list. This code can also access the unit vector from the light toward the fragment as 'lI', the unit vector from the camera toward the fragment as 'c', the outward-pointing unit normal as 'n', and dot(-lI, n) as 'lIDot'. If the light requires shadow casting, then its sampler2DShadow can be accessed as 'lIShadowSampler' and the fragment's shadow coordinates can be accessed as 'lIShadowCoords'."""
		return ("", "", "", "l" + str(lightIndex) + "u0")

class DirectionalLight(Light):
	"""A light source whose rays arrive on the scene parallel to each other. Useful for describing very distant, very bright lights such as the sun and moon. The orientation of the light in space determines the direction of the light rays; the position of the light in space has no visual effect. Directional light is never attenuated."""
	
	def __init__(self, size=0, world=None, border=(1.0, 1.0, 1.0, 1.0), isOrthogonal=True):
		"""Initializes the light. The size parameter must be 0 (indicating no shadow-casting) or a power of 2 (indicating the size of the shadow map). The isOrthogonal argument is passed along to Camera.__init__; it indicates that directional lights are orthogonal."""
		Light.__init__(self, world, isOrthogonal)
		if size <= 0:
			self.size = 0
		else:
			# Prepare a framebuffer object to facilitate shadowing.
			self.framebuffer = fbo.glGenFramebuffersEXT(1)
			fbo.glBindFramebufferEXT(fbo.GL_FRAMEBUFFER_EXT, self.framebuffer)
			# Create a shadow texture with a dummy buffer of zeros.
			self.texture = gl.glGenTextures(1)
			gl.glBindTexture(gl.GL_TEXTURE_2D, self.texture)
			gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_DEPTH_COMPONENT, size, size, 0, gl.GL_DEPTH_COMPONENT, gl.GL_FLOAT, '\x00' * (size * size * 4))
			gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
			gl.glTexParameterf(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_NEAREST)
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_DEPTH_TEXTURE_MODE, gl.GL_LUMINANCE)
			# Clamp to the border color.
			gl.glTexParameterfv(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_BORDER_COLOR, border)
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP_TO_BORDER)
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_CLAMP_TO_BORDER)
			# These comparison functions hook into the GLSL shader.
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_COMPARE_MODE, gl.GL_COMPARE_R_TO_TEXTURE)
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_COMPARE_FUNC, gl.GL_LEQUAL)
			# Attach the texture to the framebuffer object.
			gl.glDrawBuffer(gl.GL_NONE)
			gl.glReadBuffer(gl.GL_NONE)
			fbo.glFramebufferTexture2DEXT(fbo.GL_FRAMEBUFFER_EXT, fbo.GL_DEPTH_ATTACHMENT_EXT, gl.GL_TEXTURE_2D, self.texture, 0)
			# Check the well-formedness of the framebuffer object.
			status = fbo.glCheckFramebufferStatusEXT(fbo.GL_FRAMEBUFFER_EXT)
			if status == fbo.GL_FRAMEBUFFER_COMPLETE_EXT:
				self.size = size
			else:
				self.size = 0
				gl.glDeleteTextures(self.texture)
				# !! Should this be done with glDeleteBuffers() instead?
				fbo.glDeleteFramebuffersEXT(self.framebuffer)
				print 'Error: shadow map framebuffer status is', status
			# Restore the default framebuffer.
			fbo.glBindFramebufferEXT(fbo.GL_FRAMEBUFFER_EXT, 0)
	
	def __del__(self):
		"""Private. Cleans up the underlying OpenGL resources used by shadowing."""
		if self.size != 0:
			gl.glDeleteTextures(self.texture)
			# !! Should this be done with glDeleteBuffers() instead?
			fbo.glDeleteFramebuffersEXT(self.framebuffer)
		# The parent class Light has no __del__().
		#Light.__del__(self)
	
	def getCastsShadows(self):
		"""Overrides to return whether or not the light possesses a shadow map."""
		return self.size != 0
	
	def prepareShadows(self):
		"""Non-user. Overrides to prepare the shadow map, if any."""
		if self.size != 0:
			gl.glBindTexture(gl.GL_TEXTURE_2D, self.texture)
			gl.glViewport(0, 0, self.size, self.size)
			fbo.glBindFramebufferEXT(fbo.GL_FRAMEBUFFER_EXT, self.framebuffer)
			gl.glClear(gl.GL_DEPTH_BUFFER_BIT)
	
	def getShaderCode(self, lightIndex):
		"""Non-user. Overrides to return shader code expressing shadowing, if any."""
		if self.size == 0:
			return Light.getShaderCode(self, lightIndex)
		else:
			stri = str(lightIndex)
			(outers, criterion, inners, multiplier) = Light.getShaderCode(self, lightIndex)
			if criterion != "":
				criterion += " && "
			criterion += "shadow2DProj(l" + stri + "ShadowSampler, l" + stri + "ShadowCoords).r > 0.0"
			return (outers, criterion, inners, multiplier)

class SpotLight(DirectionalLight):
	"""A light source positioned somewhere in the scene, that casts a cone of light in a particular direction. The apical angle of the cone can be at most pi / 2.0 = 90.0 degrees."""
	
	def __init__(self, size=0, world=None, border=(0.0, 0.0, 0.0, 1.0), isOrthogonal=False):
		"""Initializes the light. The size parameter must be 0 (indicating no shadow-casting) or a power of 2 (indicating the size of the shadow map). The isOrthogonal argument is passed along to Camera.__init__; it indicates that spot lights are not orthogonal."""
		DirectionalLight.__init__(self, size, world, border, isOrthogonal)
		self.setConcentration(0.0)
	
	def setConcentration(self, concentration):
		"""Sets the concentration exponent of the spot light. A concentration of 0.0 produces an even distribution of light throughout the cone; a higher concentration causes the light intensity to drop off toward the boundary of the cone. The default value is 0.0."""
		self.concentration = concentration
	
	def getConcentration(self):
		"""Returns the spot light concentration exponent."""
		return self.concentration
	
	def setAngle(self, angle):
		"""Overrides so that the angle does not exceed pi / 2.0. This is the apical angle of the spot light's cone of light (not half of the apical angle)."""
		angle = min(math.pi / 2.0, angle)
		DirectionalLight.setAngle(self, angle)
		self.cosHalfAngle = math.cos(0.5 * angle)
	
	def getUniforms(self):
		"""Non-user. Overrides to return the spot direction, the cosine of the half angle, and the spot concentration exponent, in addition to the superclass' uniforms."""
		uniforms = DirectionalLight.getUniforms(self)
		return uniforms + (self.getDirection(), (self.cosHalfAngle, self.concentration))
	
	def getShaderCode(self, lightIndex):
		"""Non-user. Overrides to return shader code expressing the spot light effect."""
		# Determine where the SpotLight uniforms begin, in the tuple of uniforms.
		first = len(DirectionalLight.getUniforms(self))
		stri = str(lightIndex)
		(outers, criterion, inners, multiplier) = DirectionalLight.getShaderCode(self, lightIndex)
		# Compute the dot product of the spot direction and the direction to the fragment.
		outers += "\tfloat light" + stri + "Dot = dot(l" + stri + "u" + str(first) + ", l" + stri + ");\n"
		# Incorporate the cutoff angle into the criterion.
		if criterion != "":
			criterion += " && "
		criterion += "light" + stri + "Dot > l" + stri + "u" + str(first + 1) + "[0]"
		# In the inner statements, compute the concentration.
		inners += "\t\tfloat light" + stri + "Concentration = pow(light" + stri + "Dot, l" + stri + "u" + str(first + 1) + "[1]);\n"
		if multiplier != "":
			multiplier += " * "
		multiplier += "light" + stri + "Concentration"
		return (outers, criterion, inners, multiplier)
