# 2010 July 9 by Joshua R. Davis

import OpenGL.GL as gl
from utility import *

class Renderer2D(object):
	"""Object used by Window to render the 2D overlay layer above the 3D scene. Possesses a set of material properties. (Altering these requires the renderer to update its shader programs --- an expensive operation.) Also maintains a list of bodies (Body2D), which can be altered at any time, as long as they remain compatible with the renderer's other settings."""
	
	def __init__(self):
		"""Initializes the renderer with its default material settings."""
		# Set the default material properties.
		self.setVertexCode("modelingMatrix * vec4(a0, 0.0, 1.0)")
		self.setColorCode("vec4(1.0, 1.0, 1.0, 1.0)")
		self.setTextureCoordinateCodes(())
		self.setVaryingCodesSizes((), ())
		# Initialize other instance variables to safe values.
		self.bodies = []
		self.setWidthHeight(1.0, 1.0)
		self.drawingProgram = None
		self.pickingProgram = None
	
	def __del__(self):
		"""Non-user. Relinquishes the underlying OpenGL shader programs."""
		if self.drawingProgram != None:
			gl.glDeleteProgram(self.drawingProgram)
			self.drawingProgram = None
		if self.pickingProgram != None:
			gl.glDeleteProgram(self.pickingProgram)
			self.pickingProgram = None
		# The parent class object has no __del__().
		#object.__del__(self)
	
	def getBodies(self):
		"""Returns the renderer's list of Body2Ds, which the user can alter (append, pop, etc.) as desired, as long as they remain compatible with the rendering settings."""
		return self.bodies
	
	def setTextureCoordinateCodes(self, codes):
		"""Sets the GLSL vertex shader expressions for texture coordinates, to those given in the sequence codes. That is, codes[i] indicates where the texture coordinates for sampler i come from. For example, if codes[0] is the string 'a2', then the bodies must supply 2D texture coordinates as attribute 2. Can refer to uniform and attribute variables. The default is the empty sequence."""
		self.textureCoordinateCodes = tuple(codes)
		self.needsUpdating = True
	
	def getTextureCoordinateCodes(self):
		"""Returns the tuple of texture coordinate codes."""
		return self.textureCoordinateCodes
	
	def setVaryingCodesSizes(self, codes, sizes):
		"""Sets the GLSL vertex shader expressions for varying variables, to those given in the sequence codes, with sizes given in the sequence sizes. That is, sizes[i] indicates how many floats are in the ith varying variable and codes[i] indicates how the value of that variable is set. For example, suppose sizes[0] is 3 and codes[0] is 'a2'. Then the bodies supply 3-vectors as attribute 2; these 3-vectors are interpolated and appear in the fragment shader as v0. Can refer to uniform and attribute variables. The default for both codes and sizes is the empty sequence."""
		self.varyingCodes = tuple(codes)
		self.varyingSizes = tuple(sizes)
		self.needsUpdating = True
	
	def getVaryingCodes(self):
		"""Returns the tuple of varying variable codes."""
		return self.varyingCodes
	
	def getVaryingSizes(self):
		"""Returns the tuple of varying variable sizes."""
		return self.varyingSizes
	
	def setVertexCode(self, code):
		"""Sets the GLSL vertex shader expression for the 4D homogeneous vertex before the viewing and projection transformations. Can refer to uniform and attribute variables. The default is the string 'modelingMatrix * vec4(a0, 0.0, 1.0)', meaning that the bodies' meshes supply 2D vertices as attribute 0."""
		self.vertexCode = code
		self.needsUpdating = True
	
	def getVertexCode(self):
		"""Returns the vertex code."""
		return self.vertexCode
	
	def setColorCode(self, code):
		"""Sets the GLSL fragment shader expression for the RGBA coloration. Can refer to uniform, varying, and sampler variables. The default is the string 'vec4(1.0, 1.0, 1.0, 1.0)', meaning opaque white."""
		self.colorCode = code
		self.needsUpdating = True
	
	def getColorCode(self):
		"""Returns the color code."""
		return self.colorCode
	
	def setWidthHeight(self, width, height):
		"""Non-user method called by Window when it resizes. Informs the renderer of the window's current width and height. Subsequently notifies all bodies that the width and height have changed."""
		for body in self.bodies:
			body.setResizingWidthHeight(self.width, self.height, width, height)
		self.width = width
		self.height = height
	
	def updateAsNeeded(self):
		"""Private method. If the shader programs need updating and an example body is available, then updates the shader programs."""
		if self.needsUpdating:
			example = None
			i = 0
			while i < len(self.bodies) and example == None:
				example = self.bodies[i].getExampleBody()
				i += 1
			if example != None:
				self.updateShaderPrograms(example)
	
	def draw(self):
		"""Non-user method called by Window when it redraws. Draws the renderer's bodies."""
		self.updateAsNeeded()
		if len(self.bodies) > 0 and not self.needsUpdating:
			gl.glViewport(0, 0, self.width, self.height)
			gl.glDisable(gl.GL_DEPTH_TEST)
			gl.glEnable(gl.GL_BLEND)
			gl.glUseProgram(self.drawingProgram)
			# Establish the camera transformation.
			gl.glUniformMatrix4fv(self.drawingCameraLocation, 1, gl.GL_TRUE, (2.0 / self.width, 0.0, 0.0, -1.0, 0.0, 2.0 / self.height, 0.0, -1.0, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 1.0))
			# Draw the bodies using attribute arrays.
			for loc in self.drawingAttributeLocations:
				gl.glEnableVertexAttribArray(loc)
			for body in self.bodies:
				body.draw(self.drawingUniformLocations, self.drawingAttributeLocations, self.drawingModelingLocation)
			for loc in self.drawingAttributeLocations:
				gl.glDisableVertexAttribArray(loc)
	
	def pick(self, xy):
		"""The input xy is a screen coordinate pair (with the origin in the bottom left corner). Returns the index of the body that is topmost at that xy, or None."""
		self.updateAsNeeded()
		if len(self.bodies) == 0 or self.needsUpdating:
			return None
		else:
			gl.glDisable(gl.GL_DEPTH_TEST)
			gl.glDisable(gl.GL_BLEND)
			gl.glUseProgram(self.pickingProgram)
			# Establish the camera transformation.
			gl.glUniformMatrix4fv(self.pickingCameraLocation, 1, gl.GL_TRUE, (2.0 / self.width, 0.0, 0.0, -1.0, 0.0, 2.0 / self.height, 0.0, -1.0, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0, 0.0, 1.0))
			# Draw the bodies with prescribed colors.
			for loc in self.pickingAttributeLocations:
				if loc != -1:
					gl.glEnableVertexAttribArray(loc)
			for i in range(len(self.bodies)):
				# This hack to encode names as floats is pretty weird.
				gl.glUniform1f(self.pickingNameLocation, (i + 1) / 127.0)
				self.bodies[i].draw(self.pickingUniformLocations, self.pickingAttributeLocations, self.pickingModelingLocation)
			for loc in self.pickingAttributeLocations:
				if loc != -1:
					gl.glDisableVertexAttribArray(loc)
			# Read a 1x1 block of pixels out of the back buffer.
			gl.glReadBuffer(gl.GL_BACK)
			pixels = gl.glReadPixels(xy[0], xy[1], 1, 1, gl.GL_RGBA, gl.GL_BYTE)
			# Read this single pixel's alpha channel (which isn't multisampled).
			name = int(pixels[0][0][3])
			if name == 127:
				return None
			else:
				return name
	
	# !! handle errors from program compilation
	def updateShaderPrograms(self, body):
		"""Rewrites/compiles/links the underlying OpenGL shader programs for drawing and picking. This method is called automatically when a frame is drawn, if the rendering conditions (material codes) have changed since the previous frame. Alternatively, the user can call this method after setting the rendering conditions, if she prefers to choose the Body2D used to construct the shader programs."""
		# Discard the old shader programs.
		if self.drawingProgram != None:
			gl.glDeleteProgram(self.drawingProgram)
			self.drawingProgram = None
		if self.pickingProgram != None:
			gl.glDeleteProgram(self.pickingProgram)
			self.pickingProgram = None
		# Collect the sizes of data supplied by the body.
		uniformSizes = [len(uniform) for uniform in body.getUniforms()]
		attributeSizes = body.getMesh().getAttributeSizes()
		# Build the drawing shader program.
		vertexCode = self.getDrawingVertexShaderCode(uniformSizes, attributeSizes)
		fragmentCode = self.getDrawingFragmentShaderCode(uniformSizes, attributeSizes)
		self.drawingProgram = compiledShaderProgram(vertexCode, fragmentCode)
		# Set the drawing program's samplers before validating.
		gl.glUseProgram(self.drawingProgram)
		for i in range(len(self.textureCoordinateCodes)):
			location = gl.glGetUniformLocation(self.drawingProgram, "texture" + str(i))
			gl.glUniform1i(location, i)
		validateShaderProgram(self.drawingProgram)
		# Store the drawing program locations for future use.
		self.drawingCameraLocation = gl.glGetUniformLocation(self.drawingProgram, "cameraMatrix")
		self.drawingModelingLocation = gl.glGetUniformLocation(self.drawingProgram, "modelingMatrix")
		self.drawingUniformLocations = []
		for i in range(len(uniformSizes)):
			self.drawingUniformLocations.append(gl.glGetUniformLocation(self.drawingProgram, "u" + str(i)))
		self.drawingAttributeLocations = []
		for i in range(len(attributeSizes)):
			self.drawingAttributeLocations.append(gl.glGetAttribLocation(self.drawingProgram, "a" + str(i)))
		# Do the same steps for the picking program, in simplified form.
		vertexCode = self.getPickingVertexShaderCode(uniformSizes, attributeSizes)
		fragmentCode = self.getPickingFragmentShaderCode(uniformSizes, attributeSizes)
		self.pickingProgram = compiledShaderProgram(vertexCode, fragmentCode)
		validateShaderProgram(self.pickingProgram)
		self.pickingCameraLocation = gl.glGetUniformLocation(self.pickingProgram, "cameraMatrix")
		self.pickingModelingLocation = gl.glGetUniformLocation(self.pickingProgram, "modelingMatrix")
		self.pickingUniformLocations = []
		for i in range(len(uniformSizes)):
			self.pickingUniformLocations.append(gl.glGetUniformLocation(self.pickingProgram, "u" + str(i)))
		self.pickingAttributeLocations = []
		for i in range(len(attributeSizes)):
			self.pickingAttributeLocations.append(gl.glGetAttribLocation(self.pickingProgram, "a" + str(i)))
		# Also store the location for the picking name.
		self.pickingNameLocation = gl.glGetUniformLocation(self.pickingProgram, "pickingName")
		self.needsUpdating = False
	
	def getDrawingVertexShaderCode(self, uniformSizes, attributeSizes):
		"""Private. Returns a string containing the code for the vertex shader used in regular drawing."""
		shaderCode = ""
		# Declare the uniform variables.
		shaderCode += "uniform mat4 cameraMatrix;\n"
		shaderCode += "uniform mat4 modelingMatrix;\n"
		for i in range(len(uniformSizes)):
			if uniformSizes[i] == 1:
				shaderCode += "uniform float u" + str(i) + ";\n"
			else:
				shaderCode += "uniform vec" + str(uniformSizes[i]) + " u" + str(i) + ";\n"
		# Declare the attribute variables.
		for i in range(len(attributeSizes)):
			if attributeSizes[i] == 1:
				shaderCode += "attribute float a" + str(i) + ";\n"
			else:
				shaderCode += "attribute vec" + str(attributeSizes[i]) + " a" + str(i) + ";\n"
		# Declare the varying variables, including the texture coordinates.
		for i in range(len(self.varyingSizes)):
			if self.varyingSizes[i] == 1:
				shaderCode += "varying float v" + str(i) + ";\n"
			else:
				shaderCode += "varying vec" + str(self.varyingSizes[i]) + " v" + str(i) + ";\n"
		for i in range(len(self.textureCoordinateCodes)):
			shaderCode += "varying vec2 t" + str(i) + ";\n"
		# Begin the main body.
		shaderCode += "void main()\n{\n"
		# Copy the attribute variables, including texture coordinates, to varying variables.
		for i in range(len(self.varyingCodes)):
			stri = str(i)
			shaderCode += "\tv" + stri + " = " + self.varyingCodes[i] + ";\n"
		for i in range(len(self.textureCoordinateCodes)):
			shaderCode += "\tt" + str(i) + " = " + self.textureCoordinateCodes[i] + ";\n"
		# Finish the main body by writing the transformed vertex position.
		shaderCode += "\tgl_Position = cameraMatrix * (" + self.vertexCode + ");\n}\n"
		return shaderCode
	
	def getDrawingFragmentShaderCode(self, uniformSizes, attributeSizes):
		"""Private. Returns a string containing the code for the fragment shader used in regular drawing."""
		shaderCode = ""
		# Declare the uniform variables.
		for i in range(len(uniformSizes)):
			if uniformSizes[i] == 1:
				shaderCode += "uniform float u" + str(i) + ";\n"
			else:
				shaderCode += "uniform vec" + str(uniformSizes[i]) + " u" + str(i) + ";\n"
		# Declare the varying variables.
		for i in range(len(self.varyingSizes)):
			if self.varyingSizes[i] == 1:
				shaderCode += "varying float v" + str(i) + ";\n"
			else:
				shaderCode += "varying vec" + str(self.varyingSizes[i]) + " v" + str(i) + ";\n"
		# Declare the varying variables for texture coordinates, and the corresponding samplers and results.
		for i in range(len(self.textureCoordinateCodes)):
			stri = str(i)
			shaderCode += "varying vec2 t" + str(i) + ";\n"
			shaderCode += "uniform sampler2D texture" + stri + ";\n"
			shaderCode += "vec4 s" + stri + ";\n"
		# Begin the main body.
		shaderCode += "void main()\n{\n"
		# Get the sampled colors/properties from the texture coordinates.
		for i in range(len(self.textureCoordinateCodes)):
			stri = str(i)
			shaderCode += "\ts" + stri + " = texture2D(texture" + stri + ", t" + stri + ");\n"
		# Finish the main body by writing the fragment color.
		shaderCode += "\tgl_FragColor = " + self.colorCode + ";\n}\n"
		return shaderCode
	
	def getPickingVertexShaderCode(self, uniformSizes, attributeSizes):
		"""Private. Returns a string containing the code for the vertex shader used in drawing for picking."""
		shaderCode = ""
		# Declare the uniform variables.
		shaderCode += "uniform mat4 cameraMatrix;\n"
		shaderCode += "uniform mat4 modelingMatrix;\n"
		for i in range(len(uniformSizes)):
			if uniformSizes[i] == 1:
				shaderCode += "uniform float u" + str(i) + ";\n"
			else:
				shaderCode += "uniform vec" + str(uniformSizes[i]) + " u" + str(i) + ";\n"
		# Declare the attribute variables.
		for i in range(len(attributeSizes)):
			if attributeSizes[i] == 1:
				shaderCode += "attribute float a" + str(i) + ";\n"
			else:
				shaderCode += "attribute vec" + str(attributeSizes[i]) + " a" + str(i) + ";\n"
		# Begin the main body.
		shaderCode += "void main()\n{\n"
		# Finish the main body by writing the transformed vertex position.
		shaderCode += "\tgl_Position = cameraMatrix * (" + self.vertexCode + ");\n}\n"
		return shaderCode
	
	def getPickingFragmentShaderCode(self, uniformSizes, attributeSizes):
		"""Private. Returns a string containing the code for the fragment shader used in drawing for picking."""
		shaderCode = ""
		# Declare the uniform variables.
		for i in range(len(uniformSizes)):
			if uniformSizes[i] == 1:
				shaderCode += "uniform float u" + str(i) + ";\n"
			else:
				shaderCode += "uniform vec" + str(uniformSizes[i]) + " u" + str(i) + ";\n"
		# Declare the picking name.
		shaderCode += "uniform float pickingName;\n"
		# Begin the main body.
		shaderCode += "void main()\n{\n"
		# Finish the main body by writing the fragment color.
		shaderCode += "\tgl_FragColor = vec4(pickingName, pickingName, pickingName, pickingName);\n}\n"
		return shaderCode
