# 2010 October 5 by Joshua R. Davis

import OpenGL.GL as gl
import OpenGL.GL.EXT.framebuffer_object as fbo
from utility import *

class Renderer3D(object):
	"""Object used by Window to render 3D scenes. Possesses a camera, a set of lights, a set of lighting options, and a set of material properties. Altering these requires the renderer to update its shader programs --- an expensive operation. Also maintains a list of bodies (Body3D), which can be altered at any time, as long as they remain compatible with the renderer's other settings."""
	
	"""Basic Methods"""
	
	def __init__(self):
		"""Initializes the renderer with its defaults: no lights, only diffuse lighting, no camera, no bodies."""
		# Set the default data sources.
		self.setTextureCoordinateCodes(())
		self.setVaryingCodesSizes((), ())
		self.setVertexCode("modelingMatrix * vec4(a0, 1.0)")
		self.setNormalCode("vec3(0.0, 0.0, 1.0)")
		# Set the default lighting properties.
		self.setIsAmbient(True)
		self.setAmbientColor((0.1, 0.1, 0.1))
		self.setIsEmissive(False)
		self.setEmissiveCode("vec4(0.0, 0.0, 0.0, 1.0)")
		self.setIsDiffuse(True)
		self.setDiffuseCode("vec4(0.5, 0.5, 0.5, 1.0)")
		self.setIsSpecular(False)
		self.setSpecularCode("vec4(1.0, 1.0, 1.0, 1.0)")
		self.setShininessCode("64.0")
		self.setIsAttenuated(False)
		self.setAttenuation(0.0, 0.0, 0.01)
		# Initialize other instance variables to safe values.
		self.setLights(())
		self.setCamera(None)
		self.bodies = []
		self.drawingProgram = None
		self.pickingProgram = None
		self.width = 1
		self.height = 1
	
	def __del__(self):
		"""Non-user. Relinquishes the underlying OpenGL shader programs."""
		if self.drawingProgram != None:
			gl.glDeleteProgram(self.drawingProgram)
			self.drawingProgram = None
		if self.pickingProgram != None:
			gl.glDeleteProgram(self.pickingProgram)
			self.pickingProgram = None
		# The parent class object has no __del__().
		#object.__del__(self)
	
	def getBodies(self):
		"""Returns the renderer's list of Body3Ds, which the user can alter (append, pop, etc.) as desired, as long as they remain compatible with the rendering settings."""
		return self.bodies
	
	def setWidthHeight(self, width, height):
		"""Non-user method called by Window when it resizes. Informs the renderer of the window's current width and height."""
		self.width = width
		self.height = height
		if self.camera != None:
			self.camera.setAspectRatio(float(width) / float(height))
	
	def setCamera(self, camera):
		"""Sets the renderer's Camera object, which may be None. If there is no camera, then a default camera is installed upon the first drawing."""
		self.camera = camera
		self.needsUpdating = True
	
	def getCamera(self):
		"""Returns the renderer's Camera object, which may be None."""
		return self.camera
	
	def setLights(self, lights):
		"""Sets the renderer's lights to those in the given sequence."""
		self.lights = tuple(lights)
		self.needsUpdating = True
	
	def getLights(self):
		"""Returns the tuple of lights. To change the lights, use the setLights method."""
		return self.lights
	
	def updateAsNeeded(self):
		"""Private method. If there isn't a camera, then installs one. If the shader programs need updating and an example body is available, then updates the shader programs."""
		# Install a camera if there isn't one.
		if self.camera == None:
			self.setCamera(Camera())
		# Update the shader programs if necessary.
		if self.needsUpdating:
			example = None
			i = 0
			while i < len(self.bodies) and example == None:
				example = self.bodies[i].getExampleBody()
				i += 1
			if example != None:
				self.updateShaderPrograms(example)
	
	def pick(self, xy):
		"""The input xy is a screen coordinate pair (with the origin in the bottom left corner). Returns a pair (index, xyz). If there is no body occupying the screen point xy, then index and xyz are None. If there is a body at xy, then index is the index of the frontmost body, and xyz are the approximate world coordinates of the point on that body that is drawn at xy."""
		self.updateAsNeeded()
		if len(self.bodies) == 0 or self.needsUpdating:
			return (None, None)
		else:
			gl.glEnable(gl.GL_DEPTH_TEST)
			gl.glDisable(gl.GL_BLEND)
			gl.glUseProgram(self.pickingProgram)
			# Establish the camera transformation.
			gl.glUniformMatrix4fv(self.pickingCameraMatrixLocation, 1, gl.GL_TRUE, self.camera.getMatrix())
			# Draw the bodies with prescribed colors.
			for loc in self.pickingAttributeLocations:
				if loc != -1:
					gl.glEnableVertexAttribArray(loc)
			for i in range(len(self.bodies)):
				# This hack to encode names as floats is pretty weird.
				gl.glUniform1f(self.pickingNameLocation, (i + 1) / 127.0)
				self.bodies[i].draw(self.pickingUniformLocations, self.pickingAttributeLocations, self.pickingModelingLocation)
			for loc in self.pickingAttributeLocations:
				if loc != -1:
					gl.glDisableVertexAttribArray(loc)
			# Read a 1x1 block of pixels out of the back buffer.
			gl.glReadBuffer(gl.GL_BACK)
			pixels = gl.glReadPixels(xy[0], xy[1], 1, 1, gl.GL_RGBA, gl.GL_BYTE)
			# Read this single pixel's alpha channel (which isn't multisampled).
			name = int(pixels[0][0][3])
			if name == 127:
				# No 3D body was picked.
				return (None, None)
			else:
				# A 3D body was picked; compute its world coordinates.
				depths = gl.glReadPixels(xy[0], xy[1], 1, 1, gl.GL_DEPTH_COMPONENT, gl.GL_FLOAT)
				xyz = self.camera.unproject((xy[0], xy[1], float(depths[0][0])))
				return (name, xyz)
	
	def draw(self):
		"""Non-user method called by Window when it redraws. Updates the shader programs if necessary, and draws the bodies."""
		self.updateAsNeeded()
		if len(self.bodies) > 0 and not self.needsUpdating:
			# Let the lights construct their shadow maps.
			#gl.glPolygonOffset(1.1, 4.0)
			#gl.glEnable(gl.GL_POLYGON_OFFSET_FILL)
			gl.glCullFace(gl.GL_FRONT)
			gl.glEnable(gl.GL_DEPTH_TEST)
			gl.glDisable(gl.GL_BLEND)
			gl.glUseProgram(self.pickingProgram)
			gl.glUniform1f(self.pickingNameLocation, 0.0)
			for loc in self.pickingAttributeLocations:
				if loc != -1:
					gl.glEnableVertexAttribArray(loc)
			textureUnitIndex = len(self.textureCoordinateCodes) - 1
			for i in range(len(self.lights)):
				if self.drawingLightMatrixLocations[i] != -1:
					textureUnitIndex += 1
					gl.glActiveTexture(gl.GL_TEXTURE0 + textureUnitIndex)
					gl.glUniformMatrix4fv(self.pickingCameraMatrixLocation, 1, gl.GL_TRUE, self.lights[i].getMatrix())
					self.lights[i].prepareShadows()
					for body in self.bodies:
						body.draw(self.pickingUniformLocations, self.pickingAttributeLocations, self.pickingModelingLocation)
			# Prepare for main drawing.
			#gl.glDisable(gl.GL_POLYGON_OFFSET_FILL)
			for loc in self.pickingAttributeLocations:
				if loc != -1:
					gl.glDisableVertexAttribArray(loc)
			fbo.glBindFramebufferEXT(fbo.GL_FRAMEBUFFER_EXT, 0)
			gl.glCullFace(gl.GL_BACK)
			gl.glViewport(0, 0, self.width, self.height)
			gl.glUseProgram(self.drawingProgram)
			# Establish the camera transformation (and the camera direction for lighting, if applicable).
			gl.glUniformMatrix4fv(self.drawingCameraMatrixLocation, 1, gl.GL_TRUE, self.camera.getMatrix())
			if self.drawingCameraDirectionLocation != -1:
				if self.camera.getIsOrthogonal():
					v = self.camera.getDirection()
				else:
					v = self.camera.getBody().getPosition()
				gl.glUniform3f(self.drawingCameraDirectionLocation, v[0], v[1], v[2])
			# Set the lighting uniforms.
			if self.isAmbient:
				gl.glUniform4f(self.ambientLocation, self.ambient[0], self.ambient[1], self.ambient[2], 1.0)
			if self.isDiffuse or self.isSpecular != False:
				if self.isAttenuated:
					gl.glUniform3f(self.drawingAttenuationLocation, self.constantAttenuation, self.linearAttenuation, self.quadraticAttenuation)
				for i in range(len(self.lights)):
					if self.lights[i].getIsDirectional():
						v = self.lights[i].getDirection()
					else:
						v = self.lights[i].getBody().getPosition()
					gl.glUniform3f(self.drawingLightDirectionLocations[i], v[0], v[1], v[2])
					self.lights[i].draw(self.drawingLightUniformLocations[i])
					if self.lights[i].getCastsShadows():
						lightMatrix = self.lights[i].getMatrix()
						biasMatrix = (0.5, 0.0, 0.0, 0.5, 0.0, 0.5, 0.0, 0.5, 0.0, 0.0, 0.5, 0.5, 0.0, 0.0, 0.0, 1.0)
						matrix = matrixProduct(biasMatrix, lightMatrix, 4, 4, 4)
						gl.glUniformMatrix4fv(self.drawingLightMatrixLocations[i], 1, gl.GL_TRUE, matrix)
			# Draw the bodies using attribute arrays.
			for loc in self.drawingAttributeLocations:
				gl.glEnableVertexAttribArray(loc)
			for body in self.bodies:
				body.draw(self.drawingUniformLocations, self.drawingAttributeLocations, self.drawingModelingLocation)
			for loc in self.drawingAttributeLocations:
				gl.glDisableVertexAttribArray(loc)
	
	# !! handle errors from program compilation
	def updateShaderPrograms(self, body):
		"""Rewrites/compiles/links the underlying OpenGL shader programs for drawing and picking. This method is called automatically when a frame is drawn, if the rendering conditions (material codes, lights, camera) have changed since the previous frame. Alternatively, the user can call this method after setting the rendering conditions, if she prefers to choose the Body3D used to construct the shader programs."""
		# Discard the old shader programs.
		if self.drawingProgram != None:
			gl.glDeleteProgram(self.drawingProgram)
			self.drawingProgram = None
		if self.pickingProgram != None:
			gl.glDeleteProgram(self.pickingProgram)
			self.pickingProgram = None
		# Collect the sizes of data supplied by the body.
		uniformSizes = [len(uniform) for uniform in body.getUniforms()]
		attributeSizes = body.getMesh().getAttributeSizes()
		# Build the drawing shader program.
		vertexCode = self.getDrawingVertexShaderCode(uniformSizes, attributeSizes)
		fragmentCode = self.getDrawingFragmentShaderCode(uniformSizes, attributeSizes)
		self.drawingProgram = compiledShaderProgram(vertexCode, fragmentCode)
		# Set the drawing program's samplers before validating.
		gl.glUseProgram(self.drawingProgram)
		for i in range(len(self.textureCoordinateCodes)):
			location = gl.glGetUniformLocation(self.drawingProgram, "texture" + str(i))
			gl.glUniform1i(location, i)
		if self.isDiffuse or self.isSpecular != False:
			textureUnitIndex = len(self.textureCoordinateCodes) - 1
			for i in range(len(self.lights)):
				if self.lights[i].getCastsShadows():
					textureUnitIndex += 1
					location = gl.glGetUniformLocation(self.drawingProgram, "l" + str(i) + "ShadowSampler")
					gl.glUniform1i(location, textureUnitIndex)
		validateShaderProgram(self.drawingProgram)
		# Store the drawing program locations for future use.
		self.drawingCameraMatrixLocation = gl.glGetUniformLocation(self.drawingProgram, "cameraMatrix")
		self.drawingModelingLocation = gl.glGetUniformLocation(self.drawingProgram, "modelingMatrix")
		self.drawingUniformLocations = []
		for i in range(len(uniformSizes)):
			self.drawingUniformLocations.append(gl.glGetUniformLocation(self.drawingProgram, "u" + str(i)))
		self.drawingAttributeLocations = []
		for i in range(len(attributeSizes)):
			self.drawingAttributeLocations.append(gl.glGetAttribLocation(self.drawingProgram, "a" + str(i)))
		# Store the drawing program lighting locations.
		if self.isSpecular == False:
			self.drawingCameraDirectionLocation = -1
		elif self.camera.getIsOrthogonal():
			self.drawingCameraDirectionLocation = gl.glGetUniformLocation(self.drawingProgram, "c")
		else:
			self.drawingCameraDirectionLocation = gl.glGetUniformLocation(self.drawingProgram, "cPosition")
		if self.isAmbient:
			self.ambientLocation = gl.glGetUniformLocation(self.drawingProgram, "ambientColor")
		if self.isDiffuse or self.isSpecular != False:
			if self.isAttenuated:
				self.drawingAttenuationLocation = gl.glGetUniformLocation(self.drawingProgram, "attenuation")
			self.drawingLightDirectionLocations = []
			self.drawingLightUniformLocations = []
			self.drawingLightMatrixLocations = []
			for i in range(len(self.lights)):
				li = "l" + str(i)
				if self.lights[i].getIsDirectional():
					self.drawingLightDirectionLocations.append(gl.glGetUniformLocation(self.drawingProgram, li))
				else:
					self.drawingLightDirectionLocations.append(gl.glGetUniformLocation(self.drawingProgram, li + "Position"))
				sizes = [len(uniform) for uniform in self.lights[i].getUniforms()]
				locs = []
				for j in range(len(sizes)):
					locs.append(gl.glGetUniformLocation(self.drawingProgram, li + "u" + str(j)))
				self.drawingLightUniformLocations.append(locs)
				if self.lights[i].getCastsShadows():
					self.drawingLightMatrixLocations.append(gl.glGetUniformLocation(self.drawingProgram, li + "Matrix"))
				else:
					self.drawingLightMatrixLocations.append(-1)
		# Do the same steps for the picking program, but ignoring the texturing and lighting stuff.
		vertexCode = self.getPickingVertexShaderCode(uniformSizes, attributeSizes)
		fragmentCode = self.getPickingFragmentShaderCode(uniformSizes, attributeSizes)
		self.pickingProgram = compiledShaderProgram(vertexCode, fragmentCode)
		validateShaderProgram(self.pickingProgram)
		self.pickingCameraMatrixLocation = gl.glGetUniformLocation(self.pickingProgram, "cameraMatrix")
		self.pickingModelingLocation = gl.glGetUniformLocation(self.pickingProgram, "modelingMatrix")
		self.pickingUniformLocations = []
		for i in range(len(uniformSizes)):
			self.pickingUniformLocations.append(gl.glGetUniformLocation(self.pickingProgram, "u" + str(i)))
		self.pickingAttributeLocations = []
		for i in range(len(attributeSizes)):
			self.pickingAttributeLocations.append(gl.glGetAttribLocation(self.pickingProgram, "a" + str(i)))
		# Also store the location for the picking name.
		self.pickingNameLocation = gl.glGetUniformLocation(self.pickingProgram, "pickingName")
		self.needsUpdating = False
	
	def getDrawingVertexShaderCode(self, uniformSizes, attributeSizes):
		"""Private. Returns a string containing the code for the vertex shader used in regular drawing."""
		shaderCode = ""
		# Declare the uniform variables.
		shaderCode += "uniform mat4 cameraMatrix;\n"
		shaderCode += "uniform mat4 modelingMatrix;\n"
		for i in range(len(uniformSizes)):
			if uniformSizes[i] == 1:
				shaderCode += "uniform float u" + str(i) + ";\n"
			else:
				shaderCode += "uniform vec" + str(uniformSizes[i]) + " u" + str(i) + ";\n"
		# Declare the attribute variables.
		for i in range(len(attributeSizes)):
			if attributeSizes[i] == 1:
				shaderCode += "attribute float a" + str(i) + ";\n"
			else:
				shaderCode += "attribute vec" + str(attributeSizes[i]) + " a" + str(i) + ";\n"
		# Declare the varying variables, including the texture coordinates.
		for i in range(len(self.varyingSizes)):
			if self.varyingSizes[i] == 1:
				shaderCode += "varying float v" + str(i) + ";\n"
			else:
				shaderCode += "varying vec" + str(self.varyingSizes[i]) + " v" + str(i) + ";\n"
		for i in range(len(self.textureCoordinateCodes)):
			shaderCode += "varying vec2 t" + str(i) + ";\n"
		# Declare the special variables used in lighting.
		shaderCode += "varying vec3 v;\n"
		if self.isDiffuse or self.isSpecular != False:
			for i in range(len(self.lights)):
				if self.lights[i].getCastsShadows():
					stri = str(i)
					shaderCode += "uniform mat4 l" + stri + "Matrix;\n"
					shaderCode += "varying vec4 l" + stri + "ShadowCoords;\n"
		# Begin the main body.
		shaderCode += "void main()\n{\n"
		# Create the vertex and its varying version.
		shaderCode += "\tvec4 vertex = " + self.vertexCode + ";\n"
		shaderCode += "\tv = vertex.xyz;\n"
		# Copy the attribute variables, including texture coordinates, to varying variables.
		for i in range(len(self.varyingCodes)):
			stri = str(i)
			shaderCode += "\tv" + stri + " = " + self.varyingCodes[i] + ";\n"
		for i in range(len(self.textureCoordinateCodes)):
			shaderCode += "\tt" + str(i) + " = " + self.textureCoordinateCodes[i] + ";\n"
		# Construct shadow map coordinates for any lights that need them.
		if self.isDiffuse or self.isSpecular != False:
			for i in range(len(self.lights)):
				if self.lights[i].getCastsShadows():
					li = "l" + str(i)
					shaderCode += "\t" + li + "ShadowCoords = " + li + "Matrix * vertex;\n"
		# Finish the main body by writing the transformed vertex position.
		shaderCode += "\tgl_Position = cameraMatrix * vertex;\n}\n"
		return shaderCode
	
	def getDrawingFragmentShaderCode(self, uniformSizes, attributeSizes):
		"""Private. Returns a string containing the code for the fragment shader used in regular drawing."""
		shaderCode = ""
		# Declare the uniform variables.
		for i in range(len(uniformSizes)):
			if uniformSizes[i] == 1:
				shaderCode += "uniform float u" + str(i) + ";\n"
			else:
				shaderCode += "uniform vec" + str(uniformSizes[i]) + " u" + str(i) + ";\n"
		# Declare the varying variables.
		for i in range(len(self.varyingSizes)):
			if self.varyingSizes[i] == 1:
				shaderCode += "varying float v" + str(i) + ";\n"
			else:
				shaderCode += "varying vec" + str(self.varyingSizes[i]) + " v" + str(i) + ";\n"
		# Declare the varying variables for texture coordinates, and the corresponding samplers and results.
		for i in range(len(self.textureCoordinateCodes)):
			stri = str(i)
			shaderCode += "varying vec2 t" + str(i) + ";\n"
			shaderCode += "uniform sampler2D texture" + stri + ";\n"
			shaderCode += "vec4 s" + stri + ";\n"
		# Declare the special variables used in lighting.
		if self.isDiffuse or self.isSpecular != False:
			if self.isAttenuated:
				shaderCode += "uniform vec3 attenuation;\n"
			shaderCode += "varying vec3 v;\n"
			if self.isSpecular != False:
				if self.camera.getIsOrthogonal():
					shaderCode += "uniform vec3 c;\n"
				else:
					shaderCode += "uniform vec3 cPosition;\n"
			for i in range(len(self.lights)):
				stri = str(i)
				if self.lights[i].getIsDirectional():
					shaderCode += "uniform vec3 l" + stri + ";\n"
				else:
					shaderCode += "uniform vec3 l" + stri + "Position;\n"
				sizes = [len(uniform) for uniform in self.lights[i].getUniforms()]
				for j in range(len(sizes)):
					strj = str(j)
					if sizes[j] == 1:
						shaderCode += "uniform float l" + stri + "u" + strj + ";\n"
					else:
						shaderCode += "uniform vec" + str(sizes[j]) + " l" + stri + "u" + strj + ";\n"
				if self.lights[i].getCastsShadows():
					shaderCode += "varying vec4 l" + stri + "ShadowCoords;\n"
					shaderCode += "uniform sampler2DShadow l" + stri + "ShadowSampler;\n"
		if self.isAmbient:
			shaderCode += "uniform vec4 ambientColor;\n"
		# Begin the main body.
		shaderCode += "void main()\n{\n"
		# Get the sampled colors/properties from the texture coordinates.
		for i in range(len(self.textureCoordinateCodes)):
			stri = str(i)
			shaderCode += "\ts" + stri + " = texture2D(texture" + stri + ", t" + stri + ");\n"
		# Establish vec3 c as the unit vector from the camera to the fragment (if it wasn't set uniformly).
		if self.isSpecular != False:
			if self.camera != None and not self.camera.getIsOrthogonal():
				shaderCode += "\tvec3 c = normalize(v - cPosition);\n"
		# Establish vec3 n as the unit outward-pointing normal vector.
		shaderCode += "\tvec3 n = " + self.normalCode + ";\n"
		# Declare diffuse and specular coefficients, with the former including ambient light.
		if self.isAmbient:
			shaderCode += "\tvec4 diffuseCoefficient = ambientColor;\n"
		elif self.isDiffuse:
			shaderCode += "\tvec4 diffuseCoefficient = vec4(0.0, 0.0, 0.0, 1.0);\n"
		if self.isSpecular != False:
			shaderCode += "\tvec4 specularCoefficient = vec4(0.0, 0.0, 0.0, 1.0);\n"
			if self.shininessCode != "1.0":
				shaderCode += "\tfloat shininess = " + self.shininessCode + ";\n"
		# Let each light contribute to the diffuse and specular coefficients.
		if self.isDiffuse or self.isSpecular != False:
			for i in range(len(self.lights)):
				li = "l" + str(i)
				# Establish vec3 l? as the unit vector from light ? to the fragment (if it wasn't set uniformly).
				# Also compute attenuation, if applicable.
				if not self.lights[i].getIsDirectional():
					if self.isAttenuated:
						shaderCode += "\tvec3 " + li + "ToV = v - " + li + "Position;\n"
						shaderCode += "\tfloat " + li + "Distance = length(" + li + "ToV);\n"
						shaderCode += "\tfloat " + li + "Attenuation = 1.0 / (attenuation[0] + attenuation[1] * " + li + "Distance + attenuation[2] * " + li + "Distance * " + li + "Distance);\n"
						shaderCode += "\tvec3 " + li + " = normalize(" + li + "ToV);\n"
					else:
						shaderCode += "\tvec3 " + li + " = normalize(v - " + li + "Position);\n"
				# Compute l?Dot as the dot product of the normal and the unit vector *toward* the light.
				shaderCode += "\tfloat " + li + "Dot = -dot(" + li + ", n);\n"
				# Incorporate the light's outer shader code.
				lightCode = self.lights[i].getShaderCode(i)
				shaderCode += lightCode[0]
				# Establish the criterion for whether this light shines on this fragment.
				shaderCode += "\tif (" + li + "Dot > 0.0"
				if lightCode[1] != "":
					shaderCode += " && (" + lightCode[1] + ")"
				shaderCode += ")\n\t{\n"
				# Incorporate the light's inner shader code.
				shaderCode += lightCode[2]
				# Within the criterion, compute the diffuse reflection.
				if self.isDiffuse:
					shaderCode += "\t\tdiffuseCoefficient += "
					if self.isAttenuated:
						shaderCode += li + "Attenuation * "
					shaderCode += li + "Dot * (" + lightCode[3] + ");\n"
				# Still within the criterion, compute the specular reflection.
				if self.isSpecular != False:
					if self.isSpecular == "Phong":
						reflection = "max(0.0, -dot(c, 2.0 * " + li + "Dot * n + " + li + "))"
					else:
						reflection = "max(0.0, -dot(n, normalize(c + " + li + ")))"
					shaderCode += "\t\tspecularCoefficient += "
					if self.isAttenuated:
						shaderCode += li + "Attenuation * "
					if self.shininessCode == "1.0":
						shaderCode += reflection
					else:
						shaderCode += "pow(" + reflection + ", shininess)"
					shaderCode += " * (" + lightCode[3] + ");\n"
				# End the criterion.
				shaderCode += "\t}\n"
		# Build the final fragment color.
		finalTerms = []
		if self.isEmissive:
			finalTerms.append(self.emissiveCode)
		if self.isAmbient or self.isDiffuse:
			finalTerms.append("diffuseCoefficient * (" + self.diffuseCode + ")")
		if self.isSpecular != False:
			finalTerms.append("specularCoefficient * (" + self.specularCode + ")")
		# Finish the main body by writing the fragment color.
		shaderCode += "\tgl_FragColor = " + (" + ".join(finalTerms)) + ";\n}\n"
		return shaderCode
	
	def getPickingVertexShaderCode(self, uniformSizes, attributeSizes):
		"""Private. Returns a string containing the code for the vertex shader used in drawing for picking."""
		shaderCode = ""
		# Declare the uniform variables.
		shaderCode += "uniform mat4 cameraMatrix;\n"
		shaderCode += "uniform mat4 modelingMatrix;\n"
		for i in range(len(uniformSizes)):
			if uniformSizes[i] == 1:
				shaderCode += "uniform float u" + str(i) + ";\n"
			else:
				shaderCode += "uniform vec" + str(uniformSizes[i]) + " u" + str(i) + ";\n"
		# Declare the attribute variables.
		for i in range(len(attributeSizes)):
			if attributeSizes[i] == 1:
				shaderCode += "attribute float a" + str(i) + ";\n"
			else:
				shaderCode += "attribute vec" + str(attributeSizes[i]) + " a" + str(i) + ";\n"
		# Begin the main body.
		shaderCode += "void main()\n{\n"
		# Create the vertex and its varying version.
		shaderCode += "\tvec4 vertex = " + self.vertexCode + ";\n"
		# Finish the main body by writing the transformed vertex position.
		shaderCode += "\tgl_Position = cameraMatrix * vertex;\n}\n"
		return shaderCode
	
	def getPickingFragmentShaderCode(self, uniformSizes, attributeSizes):
		"""Private. Returns a string containing the code for the fragment shader used in drawing for picking."""
		shaderCode = ""
		# Declare the uniform variables.
		for i in range(len(uniformSizes)):
			if uniformSizes[i] == 1:
				shaderCode += "uniform float u" + str(i) + ";\n"
			else:
				shaderCode += "uniform vec" + str(uniformSizes[i]) + " u" + str(i) + ";\n"
		# Declare the picking name.
		shaderCode += "uniform float pickingName;\n"
		# Begin the main body.
		shaderCode += "void main()\n{\n"
		# Finish the main body by writing the fragment color.
		shaderCode += "\tgl_FragColor = vec4(pickingName, pickingName, pickingName, pickingName);\n}\n"
		return shaderCode
	
	"""Data Sources"""
	
	def setTextureCoordinateCodes(self, codes):
		"""Sets the GLSL vertex shader expressions for texture coordinates, to those given in the sequence codes. That is, codes[i] indicates where the texture coordinates for sampler i come from. For example, if codes[0] is the string 'a2', then the bodies must supply 2D texture coordinates as attribute 2. Can refer to uniform and attribute variables. The default is the empty sequence."""
		self.textureCoordinateCodes = tuple(codes)
		self.needsUpdating = True
	
	def getTextureCoordinateCodes(self):
		"""Returns the tuple of texture coordinate codes."""
		return self.textureCoordinateCodes
	
	def setVaryingCodesSizes(self, codes, sizes):
		"""Sets the GLSL vertex shader expressions for varying variables, to those given in the sequence codes, with sizes given in the sequence sizes. That is, sizes[i] indicates how many floats are in the ith varying variable and codes[i] indicates how the value of that variable is set. For example, suppose sizes[0] is 3 and codes[0] is 'a2'. Then the bodies supply 3-vectors as attribute 2; these 3-vectors are interpolated and appear in the fragment shader as v0. Can refer to uniform and attribute variables. The default for both codes and sizes is the empty sequence."""
		self.varyingCodes = tuple(codes)
		self.varyingSizes = tuple(sizes)
		self.needsUpdating = True
	
	def getVaryingCodes(self):
		"""Returns the tuple of varying variable codes."""
		return self.varyingCodes
	
	def getVaryingSizes(self):
		"""Returns the tuple of varying variable sizes."""
		return self.varyingSizes
	
	def setVertexCode(self, code):
		"""Sets the GLSL vertex shader expression for the vec4 world coordinates of the vertex. Can refer to uniform and attribute variables. The default is the string 'modelingMatrix * vec4(a0, 1.0)', meaning that the bodies' meshes supply 3D vertices as attribute 0."""
		self.vertexCode = code
		self.needsUpdating = True
	
	def getVertexCode(self):
		"""Returns the vertex code."""
		return self.vertexCode
	
	def setNormalCode(self, code):
		"""Sets the GLSL fragment shader expression for the vec3 world coordinates of the normal vector. Can refer to uniform, varying, and sampler variables. The default is the string 'vec3(0.0, 0.0, 1.0)', meaning the unit vector in the Z-direction. A more typical value is something like 'normalize(v0)', where varying size 0 is 3 and varying code 0 is 'normalize((modelingMatrix * vec4(a1, 0.0)).xyz)' (see setVaryingCodesSizes)."""
		self.normalCode = code
		self.needsUpdating = True
	
	def getNormalCode(self):
		"""Returns the normal code."""
		return self.normalCode
	
	"""Ambient Light"""
	
	def setIsAmbient(self, isAmbient):
		"""Sets whether the renderer should incorporate ambient light. The default is True."""
		self.isAmbient = isAmbient
		self.needsUpdating = True
	
	def getIsAmbient(self):
		"""Returns whether ambient lighting is on."""
		return self.isAmbient
	
	def setAmbientColor(self, rgb):
		"""Sets the renderer's ambient light color to the given RGB tuple. The default is (0.1, 0.1, 0.1)."""
		self.ambient = tuple(rgb)
	
	def getAmbientColor(self):
		"""Returns the renderer's ambient light color as an RGB tuple."""
		return self.ambient
	
	"""Light Emission"""
	
	def setIsEmissive(self, isEmissive):
		"""Sets whether the renderer should incorporate light emission from bodies. The default is False."""
		self.isEmissive = isEmissive
		self.needsUpdating = True
	
	def getIsEmissive(self):
		"""Returns whether light emission is on."""
		return self.isEmissive
	
	def setEmissiveCode(self, code):
		"""Sets the GLSL fragment shader expression for the material emissive light color. Can refer to uniform, varying, and sampler variables. The default is the string 'vec3(0.0, 0.0, 0.0, 1.0)', meaning that surfaces emit no light."""
		self.emissiveCode = code
		self.needsUpdating = True
	
	def getEmissiveCode(self):
		"""Returns the emissive light code."""
		return self.emissiveCode
	
	"""Diffuse Reflection"""
	
	def setIsDiffuse(self, isDiffuse):
		"""Sets whether the renderer should incorporate diffuse light reflections. The default is True."""
		self.isDiffuse = isDiffuse
		self.needsUpdating = True
	
	def getIsDiffuse(self):
		"""Returns whether diffuse reflections are on."""
		return self.isDiffuse
	
	def setDiffuseCode(self, code):
		"""Sets the GLSL fragment shader expression for the material diffuse reflection color. Can refer to uniform, varying, and sampler variables. The default is the string 'vec3(0.5, 0.5, 0.5, 1.0)', meaning that surfaces are gray."""
		self.diffuseCode = code
		self.needsUpdating = True
	
	def getDiffuseCode(self):
		"""Returns the diffuse/ambient light code."""
		return self.diffuseCode
	
	"""Specular Reflection"""
	
	def setIsSpecular(self, isSpecular):
		"""Sets whether the renderer should incorporate specular light reflections. Valid arguments are False, True, and 'Phong'. The default is False."""
		self.isSpecular = isSpecular
		self.needsUpdating = True
	
	def getIsSpecular(self):
		"""Returns the specular setting."""
		return self.specular
	
	def setSpecularCode(self, code):
		"""Sets the GLSL fragment shader expression for the material specular reflection color. Can refer to uniform, varying, and sampler variables. The default is the string 'vec4(1.0, 1.0, 1.0, 1.0)', meaning that specular highlights are bright white."""
		self.specularCode = code
		self.needsUpdating = True
	
	def getSpecularCode(self):
		"""Returns the specular light code."""
		return self.specularCode
	
	def setShininessCode(self, code):
		"""Sets the GLSL fragment shader expression for the shininess. Can refer to uniform, varying, and sampler variables. The default is the string '64.0'."""
		self.shininessCode = code
		self.needsUpdating = True
	
	def getShininessCode(self):
		"""Returns the shininess code."""
		return self.shininessCode
	
	"""Attenuation"""
	
	def setIsAttenuated(self, isAttenuated):
		"""Sets whether positional lights are attenuated. The default is False. Directional lights are never attenuated."""
		self.isAttenuated = isAttenuated
	
	def getIsAttenuated(self):
		"""Returns whether positional lights are attenuated."""
		return self.isAttenuated
	
	def setAttenuation(self, constant, linear, quadratic):
		"""Sets the coefficients to be used in light attenuation (if attenuation is turned on). The default values are 0.0, 0.0, and 0.01."""
		self.constantAttenuation = constant
		self.linearAttenuation = linear
		self.quadraticAttenuation = quadratic
	
	def getConstantAttenuation(self):
		"""Returns the constant-term coefficient used in light attenuation (if attenuation is turned on)."""
		return self.constantAttenuation
	
	def getLinearAttenuation(self):
		"""Returns the linear-term coefficient used in light attenuation (if attenuation is turned on)."""
		return self.linearAttenuation
	
	def getQuadraticAttenuation(self):
		"""Returns the quadratic-term coefficient used in light attenuation (if attenuation is turned on)."""
		return self.quadraticAttenuation
