# 2010 September 13 by Joshua R. Davis

import OpenGL.GL as gl
import OpenGL.GLUT as glut
import ode
import time
import animation

class Window(object):
	"""All programs that use the Bopagopa library must begin by instantiating a Window object. At any given time one window is current. The window handles all user events (such as keyboard and mouse) and system events (such as idling). The window keeps a Renderer3D (that handles the display of the 3D scene) and a Renderer2D (that handles the display of the 2D overlays over the 3D scene)."""
	
	"""Basic Methods"""
	
	def __init__(self, width=512, height=512, title="Bopagopa"):
		"""Initializes the window and its connection to OpenGL. Sets up a variety of defaults."""
		# Initialize GLUT.
		glut.glutInit(0)
		#glut.glutInitDisplayMode(glut.GLUT_RGBA | glut.GLUT_DOUBLE | glut.GLUT_DEPTH | glut.GLUT_MULTISAMPLE)
		glut.glutInitDisplayString("rgb double depth alpha")# samples>=4")
		glut.glutInitWindowSize(width, height)
		glut.glutInitWindowPosition(0, 0)
		self.window = glut.glutCreateWindow(title)
		# Initialize the GLUT event handlers. Idling is managed by the window status handler.
		glut.glutWindowStatusFunc(self.handleWindowStatusEvent)
		glut.glutDisplayFunc(self.handleDisplayEvent)
		glut.glutReshapeFunc(self.handleResizingEvent)
		glut.glutKeyboardFunc(self.handleKeyboardEvent)
		glut.glutSpecialFunc(self.handleKeyboardEvent)
		glut.glutMouseFunc(self.handleMouseEvent)
		glut.glutMotionFunc(self.handleMotionEvent)
		glut.glutPassiveMotionFunc(self.handlePassiveMotionEvent)
		# Set up the keyboard and mouse user interface handlers.
		self.setKeyboardHandler(None)
		self.setClickHandlerEmpty(None)
		self.setBeginDragHandlerEmpty(None)
		self.setDragHandlerEmpty(None)
		self.setEndDragHandlerEmpty(None)
		self.setClickHandler3D(None)
		self.setBeginDragHandler3D(None)
		self.setDragHandler3D(None)
		self.setEndDragHandler3D(None)
		self.setClickHandler2D(None)
		self.setBeginDragHandler2D(None)
		self.setDragHandler2D(None)
		self.setEndDragHandler2D(None)
		self.menu = None
		self.menuNames = []
		self.setMenuHandler(None, None)
		self.mouseDown = False
		self.mouse = (0, 0)
		# Make this window the global Bopagopa current window.
		setWindow(self)
		# Initialize some instance variables.
		self.world = ode.World()
		self.renderer3D = None
		self.renderer2D = None
		self.animations = []
		self.setColor((0.0, 0.0, 0.0))
		self.width = width
		self.height = height
		self.isWireframe = False
		self.isFullScreen = False
		self.modifiers = 0
		self.windowed = (glut.glutGet(glut.GLUT_WINDOW_X), glut.glutGet(glut.GLUT_WINDOW_Y), glut.glutGet(glut.GLUT_WINDOW_WIDTH), glut.glutGet(glut.GLUT_WINDOW_HEIGHT))
		# Here are some good defaults for OpenGL.
		# !! Multisampling is locked on in glutInitDisplayMode()? It disrupts picking depth, maybe picking names too.
		# glEnable(gl.GL_MULTISAMPLE)
		gl.glEnable(gl.GL_CULL_FACE)
		gl.glBlendFunc(gl.GL_SRC_ALPHA, gl.GL_ONE_MINUS_SRC_ALPHA)
	
	def __del__(self):
		"""Non-user. Destroys the associated GLUT window."""
		glut.glutDestroyWindow(self.window)
		object.__del__(self)
	
	def needsRedrawing(self):
		"""Informs the window that it needs to redraw itself on its next pass through the event loop."""
		glut.glutPostRedisplay()
	
	def getRenderer3D(self):
		"""Returns the window's 3D renderer, which may be None."""
		return self.renderer3D
	
	def setRenderer3D(self, renderer):
		"""Sets the window's 3D renderer, which may be None."""
		self.renderer3D = renderer
		if renderer != None:
			renderer.setWidthHeight(self.width, self.height)
	
	def getRenderer2D(self):
		"""Returns the window's 2D renderer, which may be None."""
		return self.renderer2D
	
	def setRenderer2D(self, renderer):
		"""Sets the window's 2D renderer, which may be None."""
		self.renderer2D = renderer
		if renderer != None:
			renderer.setWidthHeight(self.width, self.height)
	
	def getWorld(self):
		"""Returns the Open Dynamics Engine (ODE) World in which the window's 3D bodies live."""
		return self.world
	
	def getWidth(self):
		"""Returns the width of the window in pixels."""
		return self.width
	
	def getHeight(self):
		"""Returns the height of the window in pixels."""
		return self.height
	
	def setColor(self, rgb):
		"""Sets the window's background RGB color; alpha is forced to be 1.0."""
		self.background = rgb
		gl.glClearColor(rgb[0], rgb[1], rgb[2], 1.0)
	
	def getColor(self):
		"""Returns the window's background RGB color; alpha is always 1.0."""
		return self.background
	
	def setIsFullScreen(self, isFullScreen):
		"""Sets the window to full-screen mode (True) or windowed mode (False). The default is False."""
		if isFullScreen and not self.isFullScreen:
			# Switch from windowed mode to full-screen mode.
			self.isFullScreen = True
			self.windowed = (glut.glutGet(glut.GLUT_WINDOW_X), glut.glutGet(glut.GLUT_WINDOW_Y), glut.glutGet(glut.GLUT_WINDOW_WIDTH), glut.glutGet(glut.GLUT_WINDOW_HEIGHT))
			glut.glutFullScreen()
		elif not isFullScreen and self.isFullScreen:
			# Switch from full-screen mode to windowed mode.
			self.isFullScreen = False
			glut.glutPositionWindow(self.windowed[0], self.windowed[1])
			glut.glutReshapeWindow(self.windowed[2], self.windowed[3])
	
	def getIsFullScreen(self):
		"""Returns whether the window is in full-screen mode."""
		return self.isFullScreen
	
	def setIsWireframe(self, isWireframe):
		"""Sets the window to display wireframes (True) or filled polygons (False). The default is False."""
		if isWireframe:
			gl.glPolygonMode(gl.GL_FRONT, gl.GL_LINE)
		else:
			gl.glPolygonMode(gl.GL_FRONT, gl.GL_FILL)
		self.isWireframe = isWireframe
		self.needsRedrawing()
	
	def getIsWireframe(self):
		"""Returns whether the window is in wireframe mode or not."""
		return self.isWireframe
	
	"""Animation Methods"""
	
	def getAnimations(self):
		"""Returns the window's list of active Animations and Delays. The user can alter this list (append, pop, etc.) as desired."""
		return self.animations
	
	def makeDelayer(self, duration, function, arguments):
		"""Shortcut for creating a Delay and appending it to the window's animations. The delayed action takes place duration seconds in the future."""
		present = time.time()
		delayer = animation.Delayer(present + duration, function, arguments)
		self.animations.append(delayer)
	
	def makePositioner(self, duration, function, startArgs, endArgs, repetitions=1.0, rescaler=None):
		"""Shortcut for creating a PositionAnimation and appending it to the window's animations. The animation starts immediately and lasts for the given duration (in seconds)."""
		present = time.time()
		ani = animation.Positioner(present, present + duration, function, startArgs, endArgs, repetitions, rescaler)
		self.animations.append(ani)
	
	def makeDisplacer(self, duration, function, zeroArgs, displacementArgs, repetitions=1.0, rescaler=None):
		"""Shortcut for creating a DisplacementAnimation and appending it to the window's animations. The animation starts immediately and lasts for the given duration (in seconds)."""
		present = time.time()
		ani = animation.Displacer(present, present + duration, function, zeroArgs, displacementArgs, repetitions, rescaler)
		self.animations.append(ani)
	
	def trackTime(self, function):
		"""Shortcut for creating a DisplacementAnimation that tracks real time. That is, on each frame of the animation the given function (or method) is called with a single float argument: the number of seconds that have passed since the last frame of the animation. (On the first frame, this is the number of seconds that have passed since the start of the animation.)"""
		self.makeDisplacer(1.0, function, [0.0], [1.0], None)
	
	"""User Interface Methods"""
	
	def mouseIsDown(self):
		"""Returns a Boolean indicating whether the mouse button is down."""
		return self.mouseDown
	
	def getMouse(self):
		"""Returns the (x, y) integer coordinates of the mouse pointer, relative to the bottom left corner of the window."""
		return self.mouse
	
	def shiftKeyWasDown(self):
		"""Returns Boolean indicating whether the shift key was down at the time of the most recent user interface event."""
		return not (not (self.modifiers & glut.GLUT_ACTIVE_SHIFT))
	
	def controlKeyWasDown(self):
		"""Returns Boolean indicating whether the control key was down at the time of the most recent user interface event."""
		return not (not (self.modifiers & glut.GLUT_ACTIVE_CTRL))
	
	def altKeyWasDown(self):
		"""Returns Boolean indicating whether the alt key was down at the time of the most recent user interface event."""
		return not (not (self.modifiers & glut.GLUT_ACTIVE_ALT))
	
	def pick(self, xy):
		"""The input xy is a screen coordinate pair (with origin in the bottom left corner). Returns a pair (index, coords) as follows. If index is None, then no body is at (x, y). If index is not None and coords is None, then index indicates which 2D body is at (x, y), by its index in the Renderer2D's list of bodies. If index is not None and coords is not None, then index indicates which 3D body is at (x, y) and coords is the approximate XYZ world coordinates of that spot on the body. Warning: Can handle at most about 127 2D bodies and about 127 3D bodies."""
		# Clear the back buffer to white and then restore the clearing color.
		gl.glClearColor(1.0, 1.0, 1.0, 1.0)
		gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
		self.setColor(self.getColor())
		# The default expectation is that nothing was picked.
		name = None
		xyz = None
		# Try 2D bodies and then 3D bodies.
		if self.renderer2D != None:
			name = self.renderer2D.pick(xy)
		if name == None:
			if self.renderer3D != None:
				(name, xyz) = self.renderer3D.pick(xy)
		return (name, xyz)
	
	def setKeyboardHandler(self, function):
		"""Sets the keyboard handler function, which should have interface keyboardHandler(xy, key). In addition to common characters, there are special characters GLUT_KEY_LEFT, GLUT_KEY_RIGHT, GLUT_KEY_UP, GLUT_KEY_DOWN, GLUT_KEY_PAGE_UP, GLUT_KEY_PAGE_DOWN, GLUT_KEY_HOME, GLUT_KEY_END, GLUT_KEY_INSERT, GLUT_KEY_F1, ..., GLUT_KEY_F12 (which can be found in the module OpenGL.GLUT)."""
		self.keyboardHandler = function
	
	def setClickHandlerEmpty(self, function):
		"""Sets the handler function for clicks on empty space, which should have interface clickHandlerEmpty(xy)."""
		self.clickHandlerEmpty = function
	
	def setBeginDragHandlerEmpty(self, function):
		"""Sets the handler function for beginning drags on empty space, which should have interface beginDragHandlerEmpty(xyStarting, xyCurrent)."""
		self.beginDragHandlerEmpty = function
	
	def setDragHandlerEmpty(self, function):
		"""Sets the handler function for drags on empty space, which should have interface dragHandlerEmpty(xyStarting, xyPrevious, xyCurrent)."""
		self.dragHandlerEmpty = function
	
	def setEndDragHandlerEmpty(self, function):
		"""Sets the handler function for ending drags on empty space, which should have interface endDragHandlerEmpty(xyStarting, xyPrevious, xyFinal)."""
		self.endDragHandlerEmpty = function
	
	def setClickHandler2D(self, function):
		"""Sets the handler function for clicks on 2D bodies, which should have interface clickHandler2D(xy, name)."""
		self.clickHandler2D = function
	
	def setBeginDragHandler2D(self, function):
		"""Sets the handler function for beginning drags on 2D bodies, which should have interface beginDragHandler2D(xyStarting, xyCurrent, name)."""
		self.beginDragHandler2D = function
	
	def setDragHandler2D(self, function):
		"""Sets the handler function for drags on 2D bodies, which should have interface dragHandler2D(xyStarting, xyPrevious, xyCurrent, name)."""
		self.dragHandler2D = function
	
	def setEndDragHandler2D(self, function):
		"""Sets the handler function for ending drags on 2D bodies, which should have interface endDragHandler2D(xyStarting, xyPrevious, xyFinal, name)."""
		self.endDragHandler2D = function
	
	def setClickHandler3D(self, function):
		"""Sets the handler function for clicks on 3D bodies, which should have interface clickHandler3D(xy, name, xyz)."""
		self.clickHandler3D = function
	
	def setBeginDragHandler3D(self, function):
		"""Sets the handler function for beginning drags on 3D bodies, which should have interface beginDragHandler3D(xyStarting, xyCurrent, name, xyzStarting)."""
		self.beginDragHandler3D = function
	
	def setDragHandler3D(self, function):
		"""Sets the handler function for drags on 3D bodies, which should have interface dragHandler3D(xyStarting, xyPrevious, xyCurrent, name, xyzStarting)."""
		self.dragHandler3D = function
	
	def setEndDragHandler3D(self, function):
		"""Sets the handler function for ending drags on 3D bodies, which should have interface endDragHandler3D(xyStarting, xyPrevious, xyFinal, name, xyzStarting)."""
		self.endDragHandler3D = function
	
	def setMenuHandler(self, function, names):
		"""Sets the menu and menu handler function that are attached to the right mouse button. The names argument is a list of distinct strings that appear as the menu items. The handler function should take a single string --- the menu item that was selected. Passing None for either argument disables the menu."""
		if self.menu != None:
			glut.glutDestroyMenu(self.menu)
		if function == None or names == None or len(names) == 0:
			glut.glutDetachMenu(glut.GLUT_RIGHT_BUTTON)
		else:
			self.menuNames = names
			self.menuHandler = function
			self.menu = glut.glutCreateMenu(self.handleMenuEvent)
			for i in range(len(names)):
				glut.glutAddMenuEntry(names[i], i)
			glut.glutAttachMenu(glut.GLUT_RIGHT_BUTTON)
	
	def handleWindowStatusEvent(self, state):
		"""Private method called by GLUT whenever the visibility of the window changes. Activates and deactivates processing of idling events."""
		if state == glut.GLUT_HIDDEN or state == glut.GLUT_FULLY_COVERED:
			glut.glutIdleFunc(None)
		else:
			glut.glutIdleFunc(self.handleIdlingEvent)
	
	def handleIdlingEvent(self):
		"""Private method called by GLUT whenever the window is visible and no other events are pending. Gives animations a chance to run."""
		if len(self.animations) >= 1:
			newTime = time.time()
			glut.glutPostRedisplay()
			i = 0
			while i < len(self.animations):
				if self.animations[i].update(newTime):
					self.animations.pop(i)
				else:
					i += 1
	
	def handleDisplayEvent(self):
		"""Private method called by GLUT whenever the window needs redisplaying. Tells the 3D renderer and then the 2D renderer to draw."""
		gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
		if self.renderer3D != None:
			self.renderer3D.draw()
		if self.renderer2D != None:
			self.renderer2D.draw()
		# Flush all pending OpenGL calls and swap the back buffer to the front.
		glut.glutSwapBuffers()
	
	def handleMouseEvent(self, button, state, x, y):
		"""Private method called by GLUT to handle mouse button events. The origin is in the upper-left corner of the window. (This method could be expanded to handle two or three buttons.)"""
		setWindow(self)
		self.mouse = (x, self.height - y)
		if state == glut.GLUT_DOWN:
			self.mouseDown = True
			self.mouse = (x, self.height - y)
			# A click or drag has begun.
			self.xyStarting = self.mouse
			self.xyPrevious = None
			self.xyCurrent = None
			self.modifiers = glut.glutGetModifiers()
			(self.name, self.xyzStarting) = self.pick(self.mouse)
		else:
			self.mouseDown = False
			if self.xyPrevious == None:
				# A click has ended.
				if self.name == None:
					if self.clickHandlerEmpty != None:
						self.clickHandlerEmpty(self.xyStarting)
				elif self.xyzStarting == None:
					if self.clickHandler2D != None:
						self.clickHandler2D(self.xyStarting, self.name)
				else:
					if self.clickHandler3D != None:
						self.clickHandler3D(self.xyStarting, self.name, self.xyzStarting)
			else:
				# A drag has ended.
				self.xyCurrent = self.mouse
				if self.name == None:
					if self.endDragHandlerEmpty != None:
						self.endDragHandlerEmpty(self.xyStarting, self.xyPrevious, self.xyCurrent)
				elif self.xyzStarting == None:
					if self.endDragHandler2D != None:
						self.endDragHandler2D(self.xyStarting, self.xyPrevious, self.xyCurrent, self.name)
				else:
					if self.endDragHandler3D != None:
						self.endDragHandler3D(self.xyStarting, self.xyPrevious, self.xyCurrent, self.name, self.xyzStarting)
	
	def handleMotionEvent(self, x, y):
		"""Private method called by GLUT to handle mouse-moved events when a mouse button is down. The origin is in the upper-left corner of the window."""
		setWindow(self)
		self.mouse = (x, self.height - y)
		if self.xyCurrent == None:
			# Dragging has just begun.
			self.xyCurrent = (x, self.height - y)
			if self.name == None:
				if self.beginDragHandlerEmpty != None:
					self.beginDragHandlerEmpty(self.xyStarting, self.xyCurrent)
				if self.dragHandlerEmpty != None:
					self.dragHandlerEmpty(self.xyStarting, self.xyStarting, self.xyCurrent)
			elif self.xyzStarting == None:
				if self.beginDragHandler2D != None:
					self.beginDragHandler2D(self.xyStarting, self.xyCurrent, self.name)
				if self.dragHandler2D != None:
					self.dragHandler2D(self.xyStarting, self.xyStarting, self.xyCurrent, self.name)
			else:
				if self.beginDragHandler3D != None:
					self.beginDragHandler3D(self.xyStarting, self.xyCurrent, self.name, self.xyzStarting)
				if self.dragHandler3D != None:
					self.dragHandler3D(self.xyStarting, self.xyStarting, self.xyCurrent, self.name, self.xyzStarting)
		else:
			# Dragging is in progress.
			self.xyPrevious = self.xyCurrent
			self.xyCurrent = (x, self.height - y)
			if self.name == None:
				if self.dragHandlerEmpty != None:
					self.dragHandlerEmpty(self.xyStarting, self.xyPrevious, self.xyCurrent)
			elif self.xyzStarting == None:
				if self.dragHandler2D != None:
					self.dragHandler2D(self.xyStarting, self.xyPrevious, self.xyCurrent, self.name)
			else:
				if self.dragHandler3D != None:
					self.dragHandler3D(self.xyStarting, self.xyPrevious, self.xyCurrent, self.name, self.xyzStarting)
	
	def handlePassiveMotionEvent(self, x, y):
		"""Private method called by GLUT to handle mouse-moved events when the mouse buttons are up. The origin is in the upper-left corner of the window."""
		setWindow(self)
		self.mouse = (x, self.height - y)

	def handleMenuEvent(self, index):
		"""Private method called by GLUT to handle contextual menu events. The argument is the index of the menu item in the list of menu items, starting from 0. Invokes the user-supplied menu handler, passing it the name string of the menu item."""
		setWindow(self)
		self.modifiers = glut.glutGetModifiers()
		if self.menuHandler != None:
			self.menuHandler(self.menuNames[index])
		return 0
	
	def handleKeyboardEvent(self, key, x, y):
		"""Private method called by GLUT to handle keyboard (and special keyboard) events. The origin is in the upper-left corner of the window."""
		setWindow(self)
		self.modifiers = glut.glutGetModifiers()
		if self.keyboardHandler != None:
			self.keyboardHandler((x, self.height - y), key)
	
	def handleResizingEvent(self, width, height):
		"""Private method called by GLUT whenever the window is resized. Resizes the OpenGL viewport and notifies the renderers."""
		setWindow(self)
		self.width = width
		self.height = height
		if self.renderer3D != None:
			self.renderer3D.setWidthHeight(width, height)
		if self.renderer2D != None:
			self.renderer2D.setWidthHeight(width, height)

"""Getting the Current Window"""

bopagopaWindow = None

def getWindow():
	"""Returns the current Bopagopa window --- the one that's receiving user interface events."""
	global bopagopaWindow
	return bopagopaWindow

def setWindow(window):
	"""Non-user. Sets the current Bopagopa window."""
	global bopagopaWindow
	bopagopaWindow = window
