# 2010 September 4 by Joshua R. Davis

"""
This program is a version of the classic Minesweeper game. You begin in the top left corner; your goal is to reach the bottom right corner without hitting any mines. You move by clicking in any blank tile that is adjacent (vertically, horizontally, or diagonally) to a tile that you have already explored. The tiles that you have explored thus far display numbers; the number in a tile indicates how many of the eight adjacent tiles contain mines. The tiles that you have not yet explored are blank by default, but you can mark them as having a mine by shift-clicking in them; the mark prevents you from accidentally clicking them. (If you change your mind, then you can shift-click again to remove the mark.)
"""

import bopagopa.bopagopajr as jr
import random

# This parameter controls the number of tiles on the board.
boardSize = 16
# This parameter controls the fraction of the board covered by mines.
mineFraction = 0.25
# This parameter controls the size in pixels of each tile on the board.
tileSize = 32
# These parameters controls the font used to display messages.
fontPath = "/Library/Fonts/Silom.ttf"
fontSize = 24
# This is the palette of colors used to indicate 0, ..., 7 mines nearby.
#palette = ((0.5, 0.5, 1.0, 1.0), (0.5, 0.75, 1.0, 1.0), (0.5, 1.0, 1.0, 1.0), (0.5, 1.0, 0.75, 1.0), (0.5, 1.0, 0.5, 1.0), (0.75, 1.0, 0.5, 1.0), (1.0, 1.0, 0.5, 1.0), (1.0, 0.75, 0.5, 1.0))
# This is a more subdued palette.
palette = ((0.5, 0.5, 1.0, 1.0), (0.5, 0.625, 1.0, 1.0), (0.5, 0.75, 1.0, 1.0), (0.5, 0.875, 1.0, 1.0), (0.5, 1.0, 1.0, 1.0), (0.5, 1.0, 0.875, 1.0), (0.5, 1.0, 0.75, 1.0), (0.5, 1.0, 0.625, 1.0))
# This parameter controls whether adjoining cells of the same color are joined together into blobs; blobs slow things down a little.
blobs = True

class Game(object):
	"""Represents the Mines game board. Possesses a grid of tiles, each one of 0, ..., 8, 'safeUnmarked', 'safeMarked', 'mineUnmarked', 'mineMarked'."""
	
	def __init__(self, n, fraction, size):
		"""Initializes the board and its display."""
		# Make the label bodies.
		self.texts = []
		for i in range(8):
			self.texts.append(jr.makeText(str(i), fontPath, fontSize))
			self.texts[i].setColor((0.0, 0.0, 0.0, 1.0))
		self.rectangleMesh = jr.makeRectangle(size, size).getMesh()
		self.mark = jr.Body2DJr(self.rectangleMesh)
		self.mark.setColor((0.0, 0.0, 0.0, 1.0))
		self.mark.setTexture(jr.makeFileTexture("mine.png"))
		# Construct the internal grid of tiles.
		self.n = n
		self.size = size
		self.initializeTiles(fraction)
		# Create a 2D body for each board tile; this is the color layer.
		tileTexture = jr.makeFileTexture("tile.png")
		bodies = []
		for x in range(n):
			for y in range(n):
				body = jr.Body2DJr(self.rectangleMesh)
				body.setPosition((size * (x + 0.5), size * (y + 0.5)))
				body.setTexture(tileTexture)
				bodies.append(body)
		jr.getWindow().appendBody2D(jr.Group2D(bodies))
		# Create a None for each tile label.
		bodies = []
		for x in range(n):
			for y in range(n):
				bodies.append(None)
		jr.getWindow().appendBody2D(jr.Group2D(bodies))
		# Make sure the starting tile is correctly displayed.
		self.revealTile(0, self.n - 1)
	
	def initializeTiles(self, fraction):
		"""Constructs the internal representation of the board as a grid of integers and strings."""
		self.tiles = []
		if fraction <= 0.5:
			# Construct a board full of empty tiles.
			for x in range(self.n):
				self.tiles.append(["safeUnmarked"] * self.n)
		else:
			# Construct a board full of mines.
			for x in range(self.n):
				self.tiles.append(["mineUnmarked"] * self.n)
		# Prevent the starting and ending tiles from being altered.
		self.tiles[0][self.n - 1] = 0
		self.tiles[self.n - 1][0] = 0
		numMines = int(fraction * (self.n * self.n - 2))
		if fraction <= 0.5:
			# Place mines.
			numPlaced = 0
			while numPlaced < numMines:
				x = random.randint(0, self.n - 1)
				y = random.randint(0, self.n - 1)
				if self.tiles[x][y] == "safeUnmarked":
					self.tiles[x][y] = "mineUnmarked"
					numPlaced += 1
		else:
			# Place empty tiles.
			numPlaced = self.n * self.n - 2
			while numPlaced > numMines:
				x = random.randint(0, self.n - 1)
				y = random.randint(0, self.n - 1)
				if self.tiles[x][y] == "mineUnmarked":
					self.tiles[x][y] = "safeUnmarked"
					numPlaced -= 1
		# Set the starting and ending tiles.
		self.tiles[0][self.n - 1] = 0
		if self.tiles[0][self.n - 2] == "mineUnmarked":
			self.tiles[0][self.n - 1] += 1
		if self.tiles[1][self.n - 2] == "mineUnmarked":
			self.tiles[0][self.n - 1] += 1
		if self.tiles[1][self.n - 1] == "mineUnmarked":
			self.tiles[0][self.n - 1] += 1
		self.tiles[self.n - 1][0] = "safeUnmarked"
	
	def isSteppable(self, i, j):
		"""Returns True iff the given tile is adjacent to a known tile and not marked."""
		if self.tiles[i][j] != "safeMarked" and self.tiles[i][j] != "mineMarked":
			for a in range(max(0, i - 1), min(self.n - 1, i + 1) + 1):
				for b in range(max(0, j - 1), min(self.n - 1, j + 1) + 1):
					if type(self.tiles[a][b]) == int and (a != i or b != j):
						return True
		return False
	
	def getNumAdjacentMines(self, i, j):
		"""Returns the number of mines adjacent to the given tile."""
		numMines = 0
		for a in range(max(0, i - 1), min(self.n - 1, i + 1) + 1):
			for b in range(max(0, j - 1), min(self.n - 1, j + 1) + 1):
				if self.tiles[a][b] == "mineUnmarked" or self.tiles[a][b] == "mineMarked":
					numMines += 1
		return numMines
	
	def handleClick(self, xy, name):
		"""Handles mouse clicks on the board."""
		# Compute the tile that was clicked.
		i = int(xy[0] / self.size)
		j = int(xy[1] / self.size)
		window = jr.getWindow()
		if window.shiftKeyWasDown():
			# Change the mine-marking on the tile.
			if self.tiles[i][j] == "mineUnmarked":
				self.tiles[i][j] = "mineMarked"
			elif self.tiles[i][j] == "safeUnmarked":
				self.tiles[i][j] = "safeMarked"
			elif self.tiles[i][j] == "mineMarked":
				self.tiles[i][j] = "mineUnmarked"
			elif self.tiles[i][j] == "safeMarked":
				self.tiles[i][j] = "safeUnmarked"
			if type(self.tiles[i][j]) != int:
				self.updateMarking(i, j)
		elif self.isSteppable(i, j):
			if self.tiles[i][j] == "mineUnmarked":
				# The player loses.
				self.highlightErrors()
				displayMessage("You lose!")
				window.setClickHandler2D(None)
			else:
				# This is the typical case; play continues.
				self.tiles[i][j] = self.getNumAdjacentMines(i, j)
				self.revealTile(i, j)
			if i == self.n - 1 and j == 0:
				# The player wins.
				displayMessage("You win!")
				window.setClickHandler2D(None)
		# Unless it was a useless click, the window must be redrawn.
		window.needsRedrawing()
	
	def updateMarking(self, i, j):
		"""Changes the tile's color and label to reflect a mine marking or not."""
		tile = jr.getWindow().getBodies2D()[2].getBodies()[i * self.n + j]
		bodies = jr.getWindow().getBodies2D()[3].getBodies()
		if self.tiles[i][j] == "mineMarked" or self.tiles[i][j] == "safeMarked":
			# Mark the tile.
			#tile.setColor((1.0, 1.0, 1.0, 1.0))
			mark = jr.Group2D([self.mark])
			mark.setPosition((self.size * (i + 0.5), self.size * (j + 0.5)))
			bodies[i * self.n + j] = mark
		else:
			# Remove the marking.
			#tile.setColor((1.0, 1.0, 1.0, 1.0))
			bodies[i * self.n + j] = None
	
	def highlightErrors(self):
		"""Highlights all tiles that were incorrectly marked."""
		tileBodies = jr.getWindow().getBodies2D()[2].getBodies()
		labelBodies = jr.getWindow().getBodies2D()[3].getBodies()
		for i in range(len(self.tiles)):
			for j in range(len(self.tiles[0])):
				if self.tiles[i][j] == "safeMarked":
					tileBodies[i * self.n + j].setColor((1.0, 0.0, 0.0, 1.0))
					self.tiles[i][j] = self.getNumAdjacentMines(i, j)
					text = jr.Group2D([self.texts[self.tiles[i][j]]])
					text.setPosition((self.size * (i + 0.5), self.size * (j + 0.5)))
					labelBodies[i * self.n + j] = text
				elif self.tiles[i][j] == "mineUnmarked":
					tileBodies[i * self.n + j].setColor((1.0, 0.0, 0.0, 1.0))
					mark = jr.Group2D([self.mark])
					mark.setPosition((self.size * (i + 0.5), self.size * (j + 0.5)))
					labelBodies[i * self.n + j] = mark
	
	def revealTile(self, i, j):
		"""Only called on tiles of type 0, ..., 7. Sets the tile group's color and label to match the underlying tile's data. Also adds merger rectangles as necessary."""
		# Update the color.
		jr.getWindow().getBodies2D()[2].getBodies()[i * self.n + j].setColor(palette[self.tiles[i][j]])
		# Update the text.
		text = jr.Group2D([self.texts[self.tiles[i][j]]])
		text.setPosition((self.size * (i + 0.5), self.size * (j + 0.5)))
		bodies = jr.getWindow().getBodies2D()[3].getBodies()
		bodies[i * self.n + j] = text
		if blobs:
			# Prepare to add mergers; count how many tiles around each corner match this one in color.
			northeast = 1
			northwest = 1
			southwest = 1
			southeast = 1
			if i + 1 < self.n:
				if self.tiles[i + 1][j] == self.tiles[i][j]:
					southeast += 1
					northeast += 1
				if j + 1 < self.n and self.tiles[i + 1][j + 1] == self.tiles[i][j]:
					northeast += 1
			if j + 1 < self.n:
				if self.tiles[i][j + 1] == self.tiles[i][j]:
					northeast += 1
					northwest += 1
				if i - 1 >= 0 and self.tiles[i - 1][j + 1] == self.tiles[i][j]:
					northwest += 1
			if i - 1 >= 0:
				if self.tiles[i - 1][j] == self.tiles[i][j]:
					northwest += 1
					southwest += 1
				if j - 1 >= 0 and self.tiles[i - 1][j - 1] == self.tiles[i][j]:
					southwest += 1
			if j - 1 >= 0:
				if self.tiles[i][j - 1] == self.tiles[i][j]:
					southwest += 1
					southeast += 1
				if i + 1 < self.n and self.tiles[i + 1][j - 1] == self.tiles[i][j]:
					southeast += 1
			# Get information about the tile at (i, j).
			color = jr.getWindow().getBodies2D()[2].getBodies()[i * self.n + j].getColor()
			x = self.size * (i + 0.5)
			y = self.size * (j + 0.5)
			bodies = jr.getWindow().getBodies2D()[1].getBodies()
			# Add mergers to fill in the corners.
			if northeast == 3:
				bodies.append(self.makeMerger(x + 0.5 * self.size, y + 0.5 * self.size, color))
			if northwest == 3:
				bodies.append(self.makeMerger(x - 0.5 * self.size, y + 0.5 * self.size, color))
			if southwest == 3:
				bodies.append(self.makeMerger(x - 0.5 * self.size, y - 0.5 * self.size, color))
			if southeast == 3:
				bodies.append(self.makeMerger(x + 0.5 * self.size, y - 0.5 * self.size, color))
			# Add mergers to fill in the sides.
			if i + 1 < self.n and self.tiles[i + 1][j] == self.tiles[i][j] and (southeast < 3 or northeast < 3):
				bodies.append(self.makeMerger(x + 0.5 * self.size, y, color))
			if j + 1 < self.n and self.tiles[i][j + 1] == self.tiles[i][j] and (northeast < 3 or northwest < 3):
				bodies.append(self.makeMerger(x, y + 0.5 * self.size, color))
			if i - 1 >= 0 and self.tiles[i - 1][j] == self.tiles[i][j] and (northwest < 3 or southwest < 3):
				bodies.append(self.makeMerger(x - 0.5 * self.size, y, color))
			if j - 1 >= 0 and self.tiles[i][j - 1] == self.tiles[i][j] and (southwest < 3 or southeast < 3):
				bodies.append(self.makeMerger(x, y - 0.5 * self.size, color))
	
	def makeMerger(self, x, y, color):
		"""Returns a square Body2D of the given location and color."""
		body = jr.Body2DJr(self.rectangleMesh)
		body.setColor(color)
		body.setPosition((x, y))
		return body

def displayMessage(string):
	"""Displays the given message in the predetermined font"""
	global fontPath, fontSize
	text = jr.makeText(string, fontPath, fontSize)
	text.setColor((0.0, 0.0, 0.0, 1.0))
	(width, height) = text.getTexture().getWidthHeight()
	roundedRectangle = jr.makeRoundedRectangle(width + 20.0, height + 20.0, 10.0)
	roundedRectangle.setColor((1.0, 0.0, 0.0, 0.75))
	group = jr.Group2D([roundedRectangle, text])
	group.setPosition((256.0, 256.0))
	jr.getWindow().appendBody2D(group)

# If the user ran (rather than imported) this file, then run the main body.
if __name__ == "__main__":
	# Create a window to hold the mine field snugly.
	windowSize = boardSize * tileSize
	window = jr.WindowJr(windowSize, windowSize, "Mines")
	window.setDragHandler2D(None)
	# Construct the board.
	background = jr.makeRectangle(windowSize, windowSize)
	background.setPosition((0.5 * windowSize, 0.5 * windowSize))
	window.appendBody2D(background)
	window.appendBody2D(jr.Group2D([]))
	game = Game(boardSize, mineFraction, float(tileSize))
	window.setClickHandler2D(game.handleClick)
	# Run the user interface.
	jr.beginUserInterface()
