# 2010 October 3 by Joshua R. Davis

"""
This tutorial demonstrates how to rotate and revolve solids in 3D. Rotations are much more complicated in 3D than in 2D. Perhaps the easiest way to imagine a rotation is as an axis (the axis of rotation) and an angle (how much to rotate about that axis). The tutorial uses animations to rotate and revolve smoothly over time. The program displays the earth orbiting the sun, with some aspects realistic, and others (especially scale) not. One second of the simulation represents one day in reality. The sun appears as a dim gray sphere, because the only light in the scene is inside the sun. It doesn't make sense to cast light on the sun; the only choice that makes sense is to give the sun full white emissive color. But Bopagopa Jr. doesn't support emissive color; we'd have to change over to Bopagopa to get that.
"""

import math
import bopagopa.bopagopajr as jr

if __name__ == "__main__":
	# All Bopagopa programs begin by creating a window. We also set up the user interface.
	window = jr.WindowJr()
	window.setBeginDragHandler3D(None)
	window.setDragHandler3D(None)
	# Make a light at the origin; also increase the ambient light to 20% intensity.
	light = jr.Light()
	light.setPosition((0.0, 0.0, 0.0))
	window.setLights((light,))
	window.setAmbientColor((0.2, 0.2, 0.2))
	# Make the sun and earth.
	sun = jr.makeSphere(1.0)
	axes = jr.makeAxes(2.0, 0.01)
	sunGroup = jr.Group3D(axes + (sun,))
	window.appendBody3D(sunGroup)
	earth = jr.makeSphere(0.25)
	earthTexture = jr.makeFileTexture("256earth.jpg")
	earth.setTexture(earthTexture)
	axes = jr.makeAxes(0.5, 0.01)
	earthGroup = jr.Group3D(axes + (earth,))
	window.appendBody3D(earthGroup)
	earthGroup.setPosition((2.0, 0.0, 0.0))
	earthGroup.rotate(23.44 * math.pi / 180.0, (1.0, 0.0, 0.0))
	# Create a never-ending animation to revolve the earth about the sun.
	# In the first repetition it interpolates between
	#   earth.revolve(0.0, (0.0, 0.0, 1.0), (0.0, 0.0, 0.0)) and
	#   earth.revolve(2.0 * math.pi / 365.0, (0.0, 0.0, 1.0), (0.0, 0.0, 0.0)).
	# The first of those is the "zero" argument; the difference of those is the "displacement" argument.
	window.makeDisplacer(1.0, earthGroup.revolve, (0.0, (0.0, 0.0, 1.0), (0.0, 0.0, 0.0)), (2.0 * math.pi / 365.0, (0.0, 0.0, 0.0), (0.0, 0.0, 0.0)), None)
	# Create a never-ending animation to rotate the earth about its axis.
	# In the first repetition it interpolates between
	#   earth.rotate(0.0, axis) and
	#   earth.rotate(2.0 * math.pi, axis).
	axis = earthGroup.vectorToWorld((0.0, 0.0, 1.0))
	window.makeDisplacer(1.0, earthGroup.rotate, (0.0, axis), (2.0 * math.pi, (0.0, 0.0, 0.0)), None)
	# All Bopagopa programs end by running the user interface.
	jr.beginUserInterface()
